#! /usr/bin/env python

############################################################################
#  segutils.py
#
#  Copyright 2013 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 16/11/2013
# Version: 1.1
#
# History:
# Version 1.0 - Created.
# Version 1.1 - Update to be included into RSGISLib python modules tree.
#
###########################################################################
# Import shutil
import shutil
# Import the rsgislib module
import rsgislib
# Import the image utilities module from rsgislib
import rsgislib.imageutils
# Import the image calc module from rsgislib
import rsgislib.imagecalc
# Import the image segmentation module from rsgislib
import rsgislib.segmentation
# Import the image rastergis module from rsgislib
import rsgislib.rastergis
# Import the python OS module
import os
# Import the collections module
import collections
#import the gdal module
import osgeo.gdal as gdal
# Import the python JSON module
import json

def runShepherdSegmentation(inputImg, outputClumps, outputMeanImg=None, tmpath='.', gdalformat='KEA', noStats=False, noStretch=False, noDelete=False, numClusters=60, minPxls=100, distThres=100, bands=None, sampling=100, kmMaxIter=200, processInMem=False, saveProcessStats=False, imgStretchStats="", kMeansCentres="", imgStatsJSONFile=""): 
    """
Utility function to call the segmentation algorithm of Shepherd et al. (2014).

Where:

* inputImg is a string containing the name of the input file.
* outputClumps is a string containing the name of the output clump file.
* outputMeanImg is the output mean image file (clumps attributed with pixel mean from input image) - pass 'None' to skip creating.
* tmpath is a file path for intermediate files (default is current directory).
* gdalformat is a string containing the GDAL format for the output file (default = KEA).
* noStats is a bool which specifies that no image statistics and pyramids should be built for the output images (default = False)/
* noStretch is a bool which specifies that the input image bands should not be stretched (default = False).
* noDelete is a bool which specifies that the temporary images created during processing should not be deleted once processing has been completed (default = False).
* numClusters is an int which specifies the number of clusters within the KMeans clustering (default = 60).
* minPxls is an int which specifies the minimum number pixels within a segments (default = 100).
* distThres specifies the distance threshold for joining the segments (default = 100, set to large number to turn off this option).
* bands is an array providing a subset of image bands to use (default is None to use all bands).
* sampling specify the subsampling of the image for the data used within the KMeans (default = 100; 1 == no subsampling).
* kmMaxIter maximum iterations for KMeans.
* processInMem where functions allow it perform processing in memory rather than on disk.
* saveProcessStats is a bool which specifies that the image stretch stats and the kMeans centre stats should be saved along with a header.
* imgStretchStats is a string providing the file name and path for the image stretch stats (Output).
* kMeansCentres is a string providing the file name and path for the KMeans clusters centres (don't include file extension; .gmtxt will be added to the end) (Output).
* imgStatsJSONFile is a string providing the name and path of a JSON file storing the image spatial extent and imgStretchStats and kMeansCentres file paths for use by other commands (Output).

Example::

    from rsgislib.segmentation import segutils
    
    inputImg = 'jers1palsar_stack.kea'
    outputClumps = 'jers1palsar_stack_clumps_elim_final.kea'
    outputMeanImg = 'jers1palsar_stack_clumps_elim_final_mean.kea'
    
    segutils.runShepherdSegmentation(inputImg, outputClumps, outputMeanImg, minPxls=100)


    """
    
    if saveProcessStats:
        if (imgStretchStats=="") or (kMeansCentres=="") or (imgStatsJSONFile==""):
            raise rsgislib.RSGISPyException("if image stretch and kmeans centres are to be saved then all file names (imgStretchStats, kMeansCentres, imgStatsJSONFile) need to be provided.")
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    basefile = os.path.basename(inputImg)
    basename = os.path.splitext(basefile)[0]
    
    outFileExt = rsgisUtils.getFileExtension(gdalformat)
    
    createdDIR = False
    if not os.path.isdir(tmpath):
        os.makedirs(tmpath)
        createdDIR = True

    # Get data type of input image
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    rsgisUtils = rsgislib.RSGISPyUtils()
    input_datatype = rsgisUtils.getRSGISLibDataType(gdal.GetDataTypeName(gdalDS.GetRasterBand(1).DataType))
    gdalDS = None
        
    # Select Image Bands if required
    inputImgBands = inputImg
    selectBands = False
    if not bands == None:
        print("Subsetting the image bands")
        selectBands = True
        inputImgBands = os.path.join(tmpath,basename+str("_bselect")+outFileExt)
        rsgislib.imageutils.selectImageBands(inputImg, inputImgBands, gdalformat, input_datatype, bands)        
    
    # Stretch input data if required.
    segmentFile = inputImgBands
    if not noStretch:
        segmentFile = os.path.join(tmpath,basename+str("_stchd")+outFileExt)
        strchFile = os.path.join(tmpath,basename+str("_stchdonly")+outFileExt)
        strchFileOffset = os.path.join(tmpath,basename+str("_stchdonlyOff")+outFileExt)
        strchMaskFile = os.path.join(tmpath,basename+str("_stchdmaskonly")+outFileExt)
        
        print("Stretch Input Image")
        rsgislib.imageutils.stretchImage(inputImgBands, strchFile, saveProcessStats, imgStretchStats, True, False, gdalformat, rsgislib.TYPE_8INT, rsgislib.imageutils.STRETCH_LINEARSTDDEV, 2)
        
        print("Add 1 to stretched file to ensure there are no all zeros (i.e., no data) regions created.")
        rsgislib.imagecalc.imageMath(strchFile, strchFileOffset, "b1+1", gdalformat, rsgislib.TYPE_8INT)
        
        print("Create Input Image Mask.")
        ImgBand = collections.namedtuple('ImgBands', ['bandName', 'fileName', 'bandIndex'])
        bandMathBands = list()
        bandMathBands.append(ImgBand(bandName="b1", fileName=inputImgBands, bandIndex=1))
        rsgislib.imagecalc.bandMath(strchMaskFile, "b1==0?1:0", gdalformat, rsgislib.TYPE_8INT, bandMathBands)
        
        print("Mask stretched Image.")
        rsgislib.imageutils.maskImage(strchFileOffset, strchMaskFile, segmentFile, gdalformat, rsgislib.TYPE_8INT, 0, 1)
        
        if not noDelete:
            # Deleting extra files
            rsgisUtils.deleteFileWithBasename(strchFile)
            rsgisUtils.deleteFileWithBasename(strchFileOffset)
            rsgisUtils.deleteFileWithBasename(strchMaskFile)
            
    # Perform KMEANS
    print("Performing KMeans.")
    outMatrixFile = os.path.join(tmpath,basename+str("_kmeansclusters"))
    if saveProcessStats:
        outMatrixFile = kMeansCentres
    rsgislib.imagecalc.kMeansClustering(segmentFile, outMatrixFile, numClusters, kmMaxIter, sampling, True, 0.0025, rsgislib.imagecalc.INITCLUSTER_DIAGONAL_FULL_ATTACH)
    
    # Apply KMEANS
    print("Apply KMeans to image.")
    kMeansFileZones = os.path.join(tmpath,basename+str("_kmeans")+outFileExt)
    rsgislib.segmentation.labelPixelsFromClusterCentres(segmentFile, kMeansFileZones, outMatrixFile+str(".gmtxt"), True, gdalformat)
    
    # Elimininate Single Pixels
    print("Eliminate Single Pixels.")
    kMeansFileZonesNoSgls = os.path.join(tmpath,basename+str("_kmeans_nosgl")+outFileExt)
    kMeansFileZonesNoSglsTmp = os.path.join(tmpath,basename+str("_kmeans_nosglTMP")+outFileExt)
    rsgislib.segmentation.eliminateSinglePixels(segmentFile, kMeansFileZones, kMeansFileZonesNoSgls, kMeansFileZonesNoSglsTmp, gdalformat, processInMem, True)
    
    # Clump
    print("Perform clump.")
    initClumpsFile = os.path.join(tmpath,basename+str("_clumps")+outFileExt)
    rsgislib.segmentation.clump(kMeansFileZonesNoSgls, initClumpsFile, gdalformat, processInMem, 0)
    
    # Elimininate small clumps
    print("Eliminate small pixels.")
    elimClumpsFile = os.path.join(tmpath,basename+str("_clumps_elim")+outFileExt)
    rsgislib.segmentation.rmSmallClumpsStepwise(segmentFile, initClumpsFile, elimClumpsFile, gdalformat, False, "", False, processInMem, minPxls, distThres)
    
    # Relabel clumps
    print("Relabel clumps.")
    rsgislib.segmentation.relabelClumps(elimClumpsFile, outputClumps, gdalformat, processInMem)
    
    # Populate with stats if required.
    if not noStats:
        print("Calculate image statistics and build pyramids.")
        rsgislib.rastergis.populateStats(outputClumps, True, True)
    
    # Create mean image if required.
    if not (outputMeanImg == None):
        rsgislib.segmentation.meanImage(inputImg, outputClumps, outputMeanImg, gdalformat, input_datatype)
        if not noStats:
            rsgislib.imageutils.popImageStats(outputMeanImg, True, 0, True)
    
    
    if saveProcessStats:
        gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
        geotransform = gdalDS.GetGeoTransform()
        if not geotransform is None:
            xTL = geotransform[0]
            yTL = geotransform[3]
            
            xRes = geotransform[1]
            yRes = geotransform[5]
            
            width = gdalDS.RasterXSize * xRes
            if yRes < 0:
                yRes = yRes * (-1)
            height = gdalDS.RasterYSize * yRes
            xBR = xTL + width
            yBR = yTL - height
            
            xCen = xTL + (width/2)
            yCen = yBR + (height/2)
        
            sceneData = dict()
            sceneData['KCENTRES'] = kMeansCentres+str(".gmtxt")
            sceneData['STRETCHSTATS'] = imgStretchStats
            sceneData['CENTRE_PT'] = {'X':xCen, 'Y':yCen}
            sceneData['BBOX'] = {'XMIN':xTL, 'YMIN':yBR, 'XMAX':xBR, 'YMAX':yTL}
            
            with open(imgStatsJSONFile, 'w') as outfile:
                json.dump(sceneData, outfile, sort_keys=True, indent=4, separators=(',', ': '), ensure_ascii=False)

        gdalDS = None
     
    if not noDelete:
        # Deleting extra files
        if not saveProcessStats:
            rsgisUtils.deleteFileWithBasename(outMatrixFile+str(".gmtxt"))
        rsgisUtils.deleteFileWithBasename(kMeansFileZones)
        rsgisUtils.deleteFileWithBasename(kMeansFileZonesNoSgls)
        rsgisUtils.deleteFileWithBasename(kMeansFileZonesNoSglsTmp)
        rsgisUtils.deleteFileWithBasename(initClumpsFile)
        rsgisUtils.deleteFileWithBasename(elimClumpsFile)
        if selectBands:
            rsgisUtils.deleteFileWithBasename(inputImgBands)
        if not noStretch:
            rsgisUtils.deleteFileWithBasename(segmentFile)
        if createdDIR:
            shutil.rmtree(tmpath)
            

def runShepherdSegmentationPreCalcdStats(inputImg, outputClumps, kMeansCentres, imgStretchStats, outputMeanImg=None, tmpath='.', gdalformat='KEA', noStats=False, noStretch=False, noDelete=False, minPxls=100, distThres=100, bands=None, processInMem=False): 
    """
Utility function to call the segmentation algorithm of Shepherd et al. (2014) using pre-calculated stretch stats and KMeans cluster centres.

Where:

* inputImg is a string containing the name of the input file.
* outputClumps is a string containing the name of the output clump file.
* kMeansCentres is a string providing the file name and path for the KMeans clusters centres (Input)
* imgStretchStats is a string providing the file name and path for the image stretch stats (Input - not required if noStretch=True)
* outputMeanImg is the output mean image file (clumps attributed with pixel mean from input image) - pass 'None' to skip creating.
* tmpath is a file path for intermediate files (default is current directory).
* gdalformat is a string containing the GDAL format for the output file (default = KEA).
* noStats is a bool which specifies that no image statistics and pyramids should be built for the output images (default = False)/
* noStretch is a bool which specifies that the input image bands should not be stretched (default = False).
* noDelete is a bool which specifies that the temporary images created during processing should not be deleted once processing has been completed (default = False).
* minPxls is an int which specifies the minimum number pixels within a segments (default = 100).
* distThres specifies the distance threshold for joining the segments (default = 100, set to large number to turn off this option).
* bands is an array providing a subset of image bands to use (default is None to use all bands).
* sampling specify the subsampling of the image for the data used within the KMeans (default = 100; 1 == no subsampling).
* processInMem where functions allow it perform processing in memory rather than on disk.

Example::

    from rsgislib.segmentation import segutils
    
    inputImg = 'jers1palsar_stack.kea'
    outputClumps = 'jers1palsar_stack_clumps_elim_final.kea'
    outputMeanImg = 'jers1palsar_stack_clumps_elim_final_mean.kea'
    kMeansCentres = 'jers1palsar_stack_kcentres.gmtxt'
    imgStretchStats = 'jers1palsar_stack_stchstats.txt'
    
    segutils.runShepherdSegmentationPreCalcdStats(inputImg, outputClumps, kMeansCentres, imgStretchStats, outputMeanImg, minPxls=100)

    """
    
    if not noStretch:
        if (imgStretchStats=="") or (imgStretchStats==None):
            raise rsgislib.RSGISPyException("A stretch stats file must be provided")
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    basefile = os.path.basename(inputImg)
    basename = os.path.splitext(basefile)[0]
    
    outFileExt = rsgisUtils.getFileExtension(gdalformat)
    
    createdDIR = False
    if not os.path.isdir(tmpath):
        os.makedirs(tmpath)
        createdDIR = True

    # Get data type of input image
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    rsgisUtils = rsgislib.RSGISPyUtils()
    input_datatype = rsgisUtils.getRSGISLibDataType(gdal.GetDataTypeName(gdalDS.GetRasterBand(1).DataType))
    gdalDS = None
        
    # Select Image Bands if required
    inputImgBands = inputImg
    selectBands = False
    if not bands == None:
        print("Subsetting the image bands")
        selectBands = True
        inputImgBands = os.path.join(tmpath,basename+str("_bselect")+outFileExt)
        rsgislib.imageutils.selectImageBands(inputImg, inputImgBands, gdalformat, input_datatype, bands)        
    
    # Stretch input data if required.
    segmentFile = inputImgBands
    if not noStretch:
        segmentFile = os.path.join(tmpath,basename+str("_stchd")+outFileExt)
        strchFile = os.path.join(tmpath,basename+str("_stchdonly")+outFileExt)
        strchFileOffset = os.path.join(tmpath,basename+str("_stchdonlyOff")+outFileExt)
        strchMaskFile = os.path.join(tmpath,basename+str("_stchdmaskonly")+outFileExt)
        
        print("Stretch Input Image")
        rsgislib.imageutils.stretchImageWithStats(inputImgBands, strchFile, imgStretchStats, gdalformat, rsgislib.TYPE_8INT, rsgislib.imageutils.STRETCH_LINEARMINMAX, 2)
        
        print("Add 1 to stretched file to ensure there are no all zeros (i.e., no data) regions created.")
        rsgislib.imagecalc.imageMath(strchFile, strchFileOffset, "b1+1", gdalformat, rsgislib.TYPE_8INT)
        
        print("Create Input Image Mask.")
        ImgBand = collections.namedtuple('ImgBands', ['bandName', 'fileName', 'bandIndex'])
        bandMathBands = list()
        bandMathBands.append(ImgBand(bandName="b1", fileName=inputImgBands, bandIndex=1))
        rsgislib.imagecalc.bandMath(strchMaskFile, "b1==0?1:0", gdalformat, rsgislib.TYPE_8INT, bandMathBands)
        
        print("Mask stretched Image.")
        rsgislib.imageutils.maskImage(strchFileOffset, strchMaskFile, segmentFile, gdalformat, rsgislib.TYPE_8INT, 0, 1)
        
        if not noDelete:
            # Deleting extra files
            rsgisUtils.deleteFileWithBasename(strchFile)
            rsgisUtils.deleteFileWithBasename(strchFileOffset)
            rsgisUtils.deleteFileWithBasename(strchMaskFile)
    
    # Apply KMEANS
    print("Apply KMeans to image.")
    kMeansFileZones = os.path.join(tmpath,basename+str("_kmeans")+outFileExt)
    rsgislib.segmentation.labelPixelsFromClusterCentres(segmentFile, kMeansFileZones, kMeansCentres, True, gdalformat)
    
    # Elimininate Single Pixels
    print("Eliminate Single Pixels.")
    kMeansFileZonesNoSgls = os.path.join(tmpath,basename+str("_kmeans_nosgl")+outFileExt)
    kMeansFileZonesNoSglsTmp = os.path.join(tmpath,basename+str("_kmeans_nosglTMP")+outFileExt)
    rsgislib.segmentation.eliminateSinglePixels(segmentFile, kMeansFileZones, kMeansFileZonesNoSgls, kMeansFileZonesNoSglsTmp, gdalformat, processInMem, True)
    
    # Clump
    print("Perform clump.")
    initClumpsFile = os.path.join(tmpath,basename+str("_clumps")+outFileExt)
    rsgislib.segmentation.clump(kMeansFileZonesNoSgls, initClumpsFile, gdalformat, processInMem, 0)
    
    # Elimininate small clumps
    print("Eliminate small pixels.")
    elimClumpsFile = os.path.join(tmpath,basename+str("_clumps_elim")+outFileExt)
    rsgislib.segmentation.rmSmallClumpsStepwise(segmentFile, initClumpsFile, elimClumpsFile, gdalformat, False, "", False, processInMem, minPxls, distThres)
    
    # Relabel clumps
    print("Relabel clumps.")
    rsgislib.segmentation.relabelClumps(elimClumpsFile, outputClumps, gdalformat, processInMem)
    
    # Populate with stats if required.
    if not noStats:
        print("Calculate image statistics and build pyramids.")
        rsgislib.rastergis.populateStats(outputClumps, True, True)
    
    # Create mean image if required.
    if not (outputMeanImg == None):
        rsgislib.segmentation.meanImage(inputImg, outputClumps, outputMeanImg, gdalformat, input_datatype)
        if not noStats:
            rsgislib.imageutils.popImageStats(outputMeanImg, True, 0, True)
     
    if not noDelete:
        # Deleting extra files
        rsgisUtils.deleteFileWithBasename(kMeansFileZones)
        rsgisUtils.deleteFileWithBasename(kMeansFileZonesNoSgls)
        rsgisUtils.deleteFileWithBasename(kMeansFileZonesNoSglsTmp)
        rsgisUtils.deleteFileWithBasename(initClumpsFile)
        rsgisUtils.deleteFileWithBasename(elimClumpsFile)
        if selectBands:
            rsgisUtils.deleteFileWithBasename(inputImgBands)
        if not noStretch:
            rsgisUtils.deleteFileWithBasename(segmentFile)
        if createdDIR:
            shutil.rmtree(tmpath)



def runShepherdSegmentationTestNumClumps(inputImg, outputClumpsBase, outStatsFile, outputMeanImgBase=None, tmpath='.', gdalformat='KEA', noStats=False, noStretch=False, noDelete=False, numClustersStart=10, numClustersStep=10, numOfClustersSteps=10, minPxls=10, distThres=1000000, bands=None, sampling=100, kmMaxIter=200, processInMem=False, minNormV=None, maxNormV=None, minNormMI=None, maxNormMI=None): 
    """
Utility function to call the segmentation algorithm of Shepherd et al. (2014) and to test are range of 'k' within the kMeans.

Where:

* inputImg is a string containing the name of the input file
* outputClumps is a string containing the name of the output clump file
* outStatsFile is a string containing the name of the output CSV file with the image segmentation stats
* outputMeanImg is the output mean image file (clumps attributed with pixel mean from input image) - pass 'None' to skip creating.
* tmpath is a file path for intermediate files (default is current directory).
* gdalformat is a string containing the GDAL format for the output file (default is KEA)
* noStats is a bool which specifies that no image statistics and pyramids should be built for the output images.
* noStretch is a bool which specifies that the input image bands should not be stretched.
* noDelete is a book which specifies that the temporary images created during processing should not be deleted once processing has been completed.
* numClustersStart is an int which specifies the number of clusters within the KMeans clustering to start the process
* numClustersStep is an int which specifies the number of clusters within the KMeans clustering added with each step
* numOfClustersSteps is an int which specifies the number steps (i.e., tests) which are performed.
* minPxls is an int which specifies the minimum number pixels within a segments.
* distThres specifies the distance threshold for joining the segments (default is a very large value which turns off this option.).
* bands is an array providing a subset of image bands to use (default is None to use all bands)
* sampling specify the subsampling of the image for the data used within the KMeans (1 == no subsampling; default is 100)
* kmMaxIter maximum iterations for KMeans.
* processInMem where functions allow it perform processing in memory rather than on disk.
* minNormV is a floating point =None
* maxNormV=None
* minNormMI=None
* maxNormMI=None)

Example::

    from rsgislib.segmentation import segutils


    inputImg = './WV2_525N040W_20110727_TOARefl_b762_stch.kea'
    outputClumpsBase = './OptimalTests/WV2_525N040W_20110727_Clumps'
    outputMeanImgBase = './OptimalTests/WV2_525N040W_20110727_ClumpsMean'
    tmpath='./OptimalTests/tmp/'
    outStatsFile = './OptimalTests/StatsClumps.csv'

    # Will test clump values from 10 to 200 with intervals of 10.
    segutils.runShepherdSegmentationTestNumClumps(inputImg, outputClumpsBase, outStatsFile, outputMeanImgBase=outputMeanImgBase, tmpath=tmpath, noStretch=True, numClustersStart=10, numClustersStep=10, numOfClustersSteps=20, minPxls=50, minNormV=None, maxNormV=None, minNormMI=None, maxNormMI=None)


    """
    colsPrefix = 'gs'
    calcNeighbours = True
    calcNormVals = False
    if minNormV==None or  maxNormV==None or minNormMI==None or maxNormMI==None:
        minNormV = 0.0
        maxNormV = 1.0
        minNormMI = 0.0
        maxNormMI = 1.0
        calcNormVals = True
    
    
    
    outputStats = list()
    
    numClusters = numClustersStart
    for i in range(numOfClustersSteps):
        numClusters = numClustersStart + (i * numClustersStep)
        print("Processing ", numClusters)
        outputClumps = outputClumpsBase + "_c" + str(numClusters) + ".kea"
        outputMeanImg = outputMeanImgBase + "_c" + str(numClusters) + ".kea"
        
        runShepherdSegmentation(inputImg, outputClumps, outputMeanImg=outputMeanImg, tmpath=tmpath, gdalformat=gdalformat, noStats=noStats, noStretch=noStretch, noDelete=noDelete, numClusters=numClusters, minPxls=minPxls, distThres=distThres, bands=bands, sampling=sampling, kmMaxIter=kmMaxIter, processInMem=processInMem)
        
        segScores = rsgislib.rastergis.calcGlobalSegmentationScore(outputClumps, inputImg, colsPrefix, calcNeighbours, minNormV, maxNormV, minNormMI, maxNormMI)
        
        tup = (numClusters, segScores)
        outputStats.append(tup)
        
    
    numImgBands = int(len(outputStats[0][1][1])/4)
    
    
    if calcNormVals:
        minVar = 0.0
        maxVar = 0.0
        scaleVar = 0.0
        minMI = 0.0
        maxMI = 0.0
        scaleMI = 0.0
        first = True
        
        for stat in outputStats:        
            for i in range(numImgBands):
                idxVar = i*4
                idxMI = (i*4)+1
                if first:
                    minVar = stat[1][1][idxVar] # Var Min
                    maxVar = stat[1][1][idxVar] # Var Max
                    minMI = stat[1][1][idxMI] # MI Min
                    maxMI = stat[1][1][idxMI] # MI Max
                    first = False
                else:
                    if stat[1][1][idxVar] < minVar:
                        minVar = stat[1][1][idxVar] # Set Var Min
                    elif stat[1][1][idxVar] > maxVar:
                        maxVar = stat[1][1][idxVar] # Set Var Max
                        
                    if stat[1][1][idxMI] < minMI:
                        minMI = stat[1][1][idxMI] # Set MI Min
                    elif stat[1][1][idxMI] > maxMI:
                        maxMI = stat[1][1][idxMI] # Set MI Max
        
        print("Var Min: ", minVar)
        print("Var Max: ", maxVar)
        print("MI Min: ", minMI)
        print("MI Max: ", maxMI)
    
        if (maxVar - minVar) > 0:
            scaleVar = 1/(maxVar - minVar)
        else:
            scaleVar = 1
        
        if (maxMI - minMI) > 0:
            scaleMI = 1/(maxMI - minMI)
        else:
            scaleMI = 1
    
        print("Var Scale: ", scaleVar)
        print("MI Scale: ", scaleMI)
    
        for stat in outputStats:
            gScore = 0.0
            for i in range(numImgBands):
                idxVar = i*4
                idxMI = (i*4) + 1
                idxVarNorm = (i*4) + 2
                idxMINorm = (i*4) + 3
                
                stat[1][1][idxVarNorm] = (stat[1][1][idxVar] - minVar) * scaleVar
                stat[1][1][idxMINorm] = (stat[1][1][idxMI] - minMI) * scaleMI
            
                gScore = gScore + stat[1][1][idxVarNorm] + stat[1][1][idxMINorm]
            stat[1][0] = gScore
    
    fileStats = open(outStatsFile, "w")
    
    colNames = "Clusters, Overall Score"
    for i in range(numImgBands):
        colNames = colNames + ", B" + str(i+1) + "_Variance"
        colNames = colNames + ", B" + str(i+1) + "_MI"
        colNames = colNames + ", B" + str(i+1) + "_VarianceNorm"
        colNames = colNames + ", B" + str(i+1) + "_MINorm"
    colNames = colNames + "\n"
    
    fileStats.write(colNames)
    
    print("Clusters, Overall Score")
    for stat in outputStats:
        line = str(stat[0]) + ", " + "{:.9f}".format(stat[1][0])
        print(line)
        for val in stat[1][1]:
            line = line + ", " + "{:.9f}".format(val)
        line = line + "\n"
        fileStats.write(line)  
        
    fileStats.close()
    print("Complete.\n")
    
def runShepherdSegmentationTestMinObjSize(inputImg, outputClumpsBase, outStatsFile, outputMeanImgBase=None, tmpath='.', gdalformat='KEA', noStats=False, noStretch=False, noDelete=False, numClusters=100, minPxlsStart=10, minPxlsStep=5, numOfMinPxlsSteps=20, distThres=1000000, bands=None, sampling=100, kmMaxIter=200, minNormV=None, maxNormV=None, minNormMI=None, maxNormMI=None): 
    """
Utility function to call the segmentation algorithm of Shepherd et al. (2014) and to test are range of 'k' within the kMeans.

Where:

* inputImg is a string containing the name of the input file
* outputClumps is a string containing the name of the output clump file
* outStatsFile is a string containing the name of the output CSV file with the image segmentation stats
* outputMeanImg is the output mean image file (clumps attributed with pixel mean from input image) - pass 'None' to skip creating.
* tmpath is a file path for intermediate files (default is current directory).
* gdalformat is a string containing the GDAL format for the output file (default is KEA)
* noStats is a bool which specifies that no image statistics and pyramids should be built for the output images.
* noStretch is a bool which specifies that the input image bands should not be stretched.
* noDelete is a book which specifies that the temporary images created during processing should not be deleted once processing has been completed.
* numClusters is an int which specifies the number of clusters within the KMeans clustering process
* minPxlsStart is an int which specifies the minimum number pixels within a segments at the start of processing.
* minPxlsStep is an int which specifies the minimum number pixels within a segments increment each step.
* numOfMinPxlsSteps is an int which specifies the number steps (i.e., tests) which are performed.
* distThres specifies the distance threshold for joining the segments (default is a very large value which turns off this option.).
* bands is an array providing a subset of image bands to use (default is None to use all bands)
* sampling specify the subsampling of the image for the data used within the KMeans (1 == no subsampling; default is 100)
* kmMaxIter maximum iterations for KMeans.
* minNormV is a floating point =None
* maxNormV=None
* minNormMI=None
* maxNormMI=None)

Example::

    from rsgislib.segmentation import segutils


    inputImg = './WV2_525N040W_20110727_TOARefl_b762_stch.kea'
    outputClumpsBase = './OptimalTests/WV2_525N040W_20110727_MinPxl'
    outputMeanImgBase = './OptimalTests/WV2_525N040W_20110727_MinPxlMean'
    tmpath='./OptimalTests/tmp/'
    outStatsFile = './OptimalTests/StatsMinPxl.csv'

    # Will test minimum number of pixels within an object from 10 to 100 with intervals of 5.
    segutils.runShepherdSegmentationTestMinObjSize(inputImg, outputClumpsBase, outStatsFile, outputMeanImgBase=outputMeanImgBase, tmpath=tmpath, noStretch=True, numClusters=100, minPxlsStart=5, minPxlsStep=5, numOfMinPxlsSteps=20, minNormV=None, maxNormV=None, minNormMI=None, maxNormMI=None)


    """
    colsPrefix = 'gs'
    calcNeighbours = True
    calcNormVals = False
    if minNormV==None or  maxNormV==None or minNormMI==None or maxNormMI==None:
        minNormV = 0.0
        maxNormV = 1.0
        minNormMI = 0.0
        maxNormMI = 1.0
        calcNormVals = True
    
    
    
    outputStats = list()
    
    minPxls = minPxlsStart
    for i in range(numOfMinPxlsSteps):
        minPxls = minPxlsStart + (i * minPxlsStep)
        print("Processing ", minPxls)
        outputClumps = outputClumpsBase + "_mp" + str(minPxls) + ".kea"
        outputMeanImg = outputMeanImgBase + "_mp" + str(minPxls) + ".kea"
        
        runShepherdSegmentation(inputImg, outputClumps, outputMeanImg=outputMeanImg, tmpath=tmpath, gdalformat=gdalformat, noStats=noStats, noStretch=noStretch, noDelete=noDelete, numClusters=numClusters, minPxls=minPxls, distThres=distThres, bands=bands, sampling=sampling, kmMaxIter=kmMaxIter)
        
        segScores = rsgislib.rastergis.calcGlobalSegmentationScore(outputClumps, inputImg, colsPrefix, calcNeighbours, minNormV, maxNormV, minNormMI, maxNormMI)
        
        tup = (minPxls, segScores)
        outputStats.append(tup)
        
    
    numImgBands = int(len(outputStats[0][1][1])/4)
    
    
    if calcNormVals:
        minVar = 0.0
        maxVar = 0.0
        scaleVar = 0.0
        minMI = 0.0
        maxMI = 0.0
        scaleMI = 0.0
        first = True
        
        for stat in outputStats:        
            for i in range(numImgBands):
                idxVar = i*4
                idxMI = (i*4)+1
                if first:
                    minVar = stat[1][1][idxVar] # Var Min
                    maxVar = stat[1][1][idxVar] # Var Max
                    minMI = stat[1][1][idxMI] # MI Min
                    maxMI = stat[1][1][idxMI] # MI Max
                    first = False
                else:
                    if stat[1][1][idxVar] < minVar:
                        minVar = stat[1][1][idxVar] # Set Var Min
                    elif stat[1][1][idxVar] > maxVar:
                        maxVar = stat[1][1][idxVar] # Set Var Max
                        
                    if stat[1][1][idxMI] < minMI:
                        minMI = stat[1][1][idxMI] # Set MI Min
                    elif stat[1][1][idxMI] > maxMI:
                        maxMI = stat[1][1][idxMI] # Set MI Max
        
        print("Var Min: ", minVar)
        print("Var Max: ", maxVar)
        print("MI Min: ", minMI)
        print("MI Max: ", maxMI)
    
        if (maxVar - minVar) > 0:
            scaleVar = 1/(maxVar - minVar)
        else:
            scaleVar = 1
        
        if (maxMI - minMI) > 0:
            scaleMI = 1/(maxMI - minMI)
        else:
            scaleMI = 1
    
        print("Var Scale: ", scaleVar)
        print("MI Scale: ", scaleMI)
    
        for stat in outputStats:
            gScore = 0.0
            for i in range(numImgBands):
                idxVar = i*4
                idxMI = (i*4) + 1
                idxVarNorm = (i*4) + 2
                idxMINorm = (i*4) + 3
                
                stat[1][1][idxVarNorm] = (stat[1][1][idxVar] - minVar) * scaleVar
                stat[1][1][idxMINorm] = (stat[1][1][idxMI] - minMI) * scaleMI
            
                gScore = gScore + stat[1][1][idxVarNorm] + stat[1][1][idxMINorm]
            stat[1][0] = gScore
    
    fileStats = open(outStatsFile, "w")
    
    colNames = "MinNumPxls, Overall Score"
    for i in range(numImgBands):
        colNames = colNames + ", B" + str(i+1) + "_Variance"
        colNames = colNames + ", B" + str(i+1) + "_MI"
        colNames = colNames + ", B" + str(i+1) + "_VarianceNorm"
        colNames = colNames + ", B" + str(i+1) + "_MINorm"
    colNames = colNames + "\n"
    
    fileStats.write(colNames)
    
    print("Clusters, Overall Score")
    for stat in outputStats:
        line = str(stat[0]) + ", " + "{:.9f}".format(stat[1][0])
        print(line)
        for val in stat[1][1]:
            line = line + ", " + "{:.9f}".format(val)
        line = line + "\n"
        fileStats.write(line)  
        
    fileStats.close()
    print("Complete.\n")




 #! /usr/bin/env python

############################################################################
# Copyright (c) 2012 Dr. Peter Bunting
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  A class to plot the summary files from the raster gis tools in
#           in RSGISLib.
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 04/04/2012
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
#############################################################################

import os.path
import sys
from datetime import datetime,timedelta
from math import *
import optparse
import matplotlib.pyplot as plt
import matplotlib.dates as matdate

class PlotAverageData (object):

    def readCSVFile(self, inFile, minList, maxList, meanList, medianList, stddevList, meanSDP1List, meanSDM1List):
        line = 0
        mean = 0
        stddev = 0
        for eachLine in inFile:
            eachLine = eachLine.strip()
            if eachLine != "":
                split = eachLine.split(',', eachLine.count(','))
                if line == 0:
                    minList.append(float(split[1]))
                if line == 1:
                    maxList.append(float(split[1]))
                if line == 2:
                    mean = float(split[1])
                    meanList.append(mean)
                if line == 3:
                    stddev = float(split[1])
                    stddevList.append(stddev)
                    meanSDP1List.append(mean+stddev)
                    meanSDM1List.append(mean-stddev)
                if line == 5:
                    medianList.append(float(split[1]))
                line = line + 1


    def plotData(self, outputFile, titleText, xLabelText, yLabelText, yVals, minList, maxList, meanList, medianList, meanSDP1List, meanSDM1List, plotMean, plotMedian, plotStdDev, plotRange):
        fig = plt.figure(figsize=(15, 10), dpi=120)
        ax = fig.add_subplot(111)

        if plotMean:
            ax.plot(yVals, meanList, 'k-', label='Mean', zorder=10)
        if plotMedian:
            if plotMean:
                ax.plot(yVals, medianList, 'k--', label='Median', zorder=9)
            else:
                ax.plot(yVals, medianList, 'k-', label='Median', zorder=10)
        if plotRange:
            ax.fill_between(yVals, minList, maxList, alpha=0.2, linewidth=1.0, facecolor='b', edgecolor=[0.70,0.70,0.70], label='Range', zorder=-1)
        if plotStdDev:
            ax.fill_between(yVals, meanSDP1List, meanSDM1List, alpha=0.4, linewidth=1.0, facecolor='g', edgecolor=[0.70,0.70,0.70], label='1 Std Dev', zorder=1)

        axRange = plt.axis('tight')

        plt.grid(color='k', linestyle='--', linewidth=0.5)
        plt.title(titleText)
        plt.xlabel(xLabelText)
        plt.ylabel(yLabelText)

        plt.savefig(outputFile, format='PDF')

    def run(self, cmdargs):
        print('rsgislibcmdxml.py script generates the XML commands for the ')
        print(' software library from a user defined template.')
        print('This script was distributed with RSGISLib 3.6.14')
        print('Copyright (C) 2014 Peter Bunting and Daniel Clewley')
        print('For maintainance email rsgislib-support@googlegroups.com')
        yDataValsList = list()
        minList = list()
        maxList = list()
        meanList = list()
        medianList = list()
        stddevList = list()
        meanSDP1List = list()
        meanSDM1List = list()

        for strVal in cmdargs.yDataStrVals:
            yDataValsList.append(float(strVal.strip()))

        for fileName in cmdargs.inputs:
            try:
                inputTextFile = open(fileName.strip(), 'r')
                self.readCSVFile(inputTextFile, minList, maxList, meanList, medianList, stddevList, meanSDP1List, meanSDM1List)
                inputTextFile.close()
            except IOError, e:
                print('\nCould not open file:\n', e)
                return

        if len(minList) != len(yDataValsList):
            print("Y data values not the same len as min vals.")
            return

        if len(maxList) != len(yDataValsList):
            print("Y data values not the same len as max vals.")
            return

        if len(meanList) != len(yDataValsList):
            print("Y data values not the same len as mean vals.")
            return

        if len(medianList) != len(yDataValsList):
            print("Y data values not the same len as median vals.")
            return

        if len(stddevList) != len(yDataValsList):
            print("Y data values not the same len as Std Dev vals.")
            return

        if len(meanSDP1List) != len(yDataValsList):
            print("Y data values not the same len as plus 1 std dev vals.")
            return

        if len(meanSDM1List) != len(yDataValsList):
            print("Y data values not the same len as minus 1 std dev vals.")
            return

        self.plotData(cmdargs.outputFile, cmdargs.title, cmdargs.xAxis, cmdargs.yAxis, yDataValsList, minList, maxList, meanList, medianList, meanSDP1List, meanSDM1List, cmdargs.mean, cmdargs.median, cmdargs.stddev, cmdargs.range)


# Command arguments
class CmdArgs:
  def __init__(self):
    p = optparse.OptionParser()
    p.add_option("-o","--output", dest="outputFile", default=None, help="Output PDF file.")
    p.add_option("-t","--title", dest="title", default="Title", help="Plot title.")
    p.add_option("-x","--xaxis", dest="xAxis", default="X Axis", help="Plot x-axis label.")
    p.add_option("-y","--yaxis", dest="yAxis", default="Y Axis", help="Plot y-axis label.")
    p.add_option("", "--mean", action="store_true", dest="mean", default=False, help="Plot mean values")
    p.add_option("", "--median", action="store_true", dest="median", default=False, help="Plot median values")
    p.add_option("", "--stddev", action="store_true", dest="stddev", default=False, help="Plot standard deviation values")
    p.add_option("", "--range", action="store_true", dest="range", default=False, help="Plot range values")
    p.add_option("-i","--input", action="append", dest="inputs", default=None, help="Input files in order they will be plotted.")
    p.add_option("-d","--yData", action="append", dest="yDataStrVals", default=None, help="Values for along the y-axis.")



    (options, args) = p.parse_args()
    self.__dict__.update(options.__dict__)

    if self.outputFile is None:
        p.print_help()
        print("output filename must be set.")
        sys.exit()


if __name__ == '__main__':
    cmdargs = CmdArgs()
    obj = PlotAverageData()
    obj.run(cmdargs)




import rsgislib
import rsgislib.imageutils
import rsgislib.imagecalc
import rsgislib.rastergis

from osgeo import gdal

import numpy

import skimage.segmentation

import os.path


def performFelsenszwalbSegmentation(inputImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100):
    """
A function to perform the Felsenszwalb segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

* inputImg - input image file.
* outputImg - output image file.
* gdalformat - output image file format.
* tmpDIR - temp DIR used to output PCA files
* calcStats - calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
* usePCA - if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
* nPCABands - the number of principle components outputs from the PCA - needs to be either 1 or 3.

"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')    
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
    
    segResult = skimage.segmentation.felzenszwalb(pxlVals)
    print(segResult.shape)
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)
        
        
def performQuickshiftSegmentation(inputImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, pcaPxlSample=100):
    """
A function to perform the quickshift segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

* inputImg - input image file.
* outputImg - output image file.
* gdalformat - output image file format.
* tmpDIR - temp DIR used to output PCA files
* calcStats - calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
* usePCA - if there are not 3 image bands in the input file then you can use PCA to reduce the number of image bands.
"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')
    
    if not (gdalDS.RasterCount == 3):
        gdalDS = None
        raise Exception('Input image should have three bands (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
    
    segResult = skimage.segmentation.quickshift(pxlVals)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)

def performSlicSegmentation(inputImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100):
    """
A function to perform the slic segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

* inputImg - input image file.
* outputImg - output image file.
* gdalformat - output image file format.
* tmpDIR - temp DIR used to output PCA files
* calcStats - calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
* usePCA - if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
* nPCABands - the number of principle components outputs from the PCA - needs to be either 1 or 3.
"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')   
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    multichannel = False
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel = True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel = False
    
    segResult = skimage.segmentation.slic(pxlVals, multichannel=multichannel)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)


def performWatershedSegmentation(inputImg, markersImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100):
    """
A function to perform the watershed segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

* inputImg - input image file.
* markersImg - input markers image file.
* outputImg - output image file.
* gdalformat - output image file format.
* tmpDIR - temp DIR used to output PCA files
* calcStats - calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
* usePCA - if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
* nPCABands - the number of principle components outputs from the PCA - needs to be either 1 or 3.
"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
            
    
    gdalMarkersDS = gdal.Open(markersImg, gdal.GA_ReadOnly)
    if gdalMarkersDS.RasterCount != 1:
        gdalMarkersDS = None
        raise Exception('Markers image should only have one image band.')
    
    gdalMarkerBand = gdalMarkersDS.GetRasterBand(1)
    markerArr = gdalMarkerBand.ReadAsArray()
    
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')    
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
    
    segResult = skimage.segmentation.watershed(pxlVals, markerArr)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)


def performRandomWalkerSegmentation(inputImg, markersImg, outputImg, gdalformat='KEA', noDataVal=0, tmpDIR='./tmp', calcStats=True, usePCA=False, nPCABands=3, pcaPxlSample=100):
    """
A function to perform the random walker segmentation algorithm from the
scikit-image library (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

* inputImg - input image file.
* markersImg - input markers image file - markers must be uniquely numbered.
* outputImg - output image file.
* gdalformat - output image file format.
* tmpDIR - temp DIR used to output PCA files
* calcStats - calculate image pixel statistics, histogram and image pyramids - note if you are not using a KEA file then the format needs to support RATs for this option as histogram and colour table are written to RAT.
* usePCA - if there are not 1 or 3 image bands in the input file then you can use PCA to reduce the number of image bands.
* nPCABands - the number of principle components outputs from the PCA - needs to be either 1 or 3.
"""
    # Create output image
    rsgislib.imageutils.createCopyImage(inputImg, outputImg, 1, 0, gdalformat, rsgislib.TYPE_32UINT)
    
    gdalMarkersDS = gdal.Open(markersImg, gdal.GA_ReadOnly)
    if gdalMarkersDS.RasterCount != 1:
        gdalMarkersDS = None
        raise Exception('Markers image should only have one image band.')
    
    gdalMarkerBand = gdalMarkersDS.GetRasterBand(1)
    markerArr = gdalMarkerBand.ReadAsArray()
    
    
    gdalDS = gdal.Open(inputImg, gdal.GA_ReadOnly)
    pcaImg = ''
    if usePCA and (nPCABands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 1, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA and (nPCABands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        rsgisUtils = rsgislib.RSGISPyUtils()
        inImgBaseName = os.path.splitext(os.path.basename(inputImg))[0]
        pcaImg = os.path.join(tmpDIR, inImgBaseName+'_pca.kea')
        pcaEigenValsFile = os.path.join(tmpDIR, inImgBaseName+'_eigen.mtxt')
        inDType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
        rsgislib.imagecalc.performImagePCA(inputImg, pcaImg, pcaEigenValsFile, 3, pcaPxlSample, 'KEA', inDType, noDataVal, False)
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif usePCA:
        gdalDS = None
        raise Exception('In the number of bands given the number of principle components.')  
    
    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise Exception('Input image should have either a single band or three (RGB). Consider performing PCA to reduce.')
    
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand+1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
    
    segResult = skimage.segmentation.random_walker(pxlVals, markerArr)
    segResult = segResult + 1
    
    gdalOutDS = gdal.Open(outputImg, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None
    
    if calcStats:
        rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)




#!/usr/bin/env python

"""
Utility for calculating the least cost path across a raster.

"""
############################################################################
#  leastcostpath.py
#
#  Copyright 2017 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Calculate the least cost path across a raster.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 17/08/2017
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import gdal, osr
import skimage.graph
import numpy



def _coord2pixelOffset(rasterfn, x, y):
    raster = gdal.Open(rasterfn)
    geotransform = raster.GetGeoTransform()
    originX = geotransform[0]
    originY = geotransform[3]
    pixelWidth = geotransform[1]
    pixelHeight = geotransform[5]
    xOffset = int((x - originX)/pixelWidth)
    yOffset = int((y - originY)/pixelHeight)
    return xOffset,yOffset

def _createPath(CostSurfaceImg, costSurfaceArray, startCoord, stopCoord):

    # coordinates to array index
    startCoordX = startCoord[0]
    startCoordY = startCoord[1]
    startIndexX,startIndexY = _coord2pixelOffset(CostSurfaceImg, startCoordX,startCoordY)

    stopCoordX = stopCoord[0]
    stopCoordY = stopCoord[1]
    stopIndexX,stopIndexY = _coord2pixelOffset(CostSurfaceImg, stopCoordX,stopCoordY)

    # create path
    indices, weight = skimage.graph.route_through_array(costSurfaceArray, (startIndexY,startIndexX), (stopIndexY,stopIndexX), geometric=True, fully_connected=True)
    indices = numpy.array(indices).T
    path = numpy.zeros_like(costSurfaceArray)
    path[indices[0], indices[1]] = 1
    return path

def _array2raster(newRasterfn, rasterfn, array, outFormat):
    raster = gdal.Open(rasterfn)
    geotransform = raster.GetGeoTransform()
    originX = geotransform[0]
    originY = geotransform[3]
    pixelWidth = geotransform[1]
    pixelHeight = geotransform[5]
    cols = array.shape[1]
    rows = array.shape[0]

    driver = gdal.GetDriverByName(outFormat)
    outRaster = driver.Create(newRasterfn, cols, rows, 1, gdal.GDT_Byte)
    outRaster.SetGeoTransform((originX, pixelWidth, 0, originY, 0, pixelHeight))
    outband = outRaster.GetRasterBand(1)
    outband.WriteArray(array)
    outRasterSRS = osr.SpatialReference()
    outRasterSRS.ImportFromWkt(raster.GetProjectionRef())
    outRaster.SetProjection(outRasterSRS.ExportToWkt())
    outband.FlushCache()
    outRaster = None
    raster = None

def performLeastCostPathCalc(costSurfaceImg, outputPathImg, startCoord, stopCoord, gdalformat="KEA", costImgBand=1):
    """
    Calculates least cost path for a raster surface from start coord to stop coord:

    Version of code from: https://pcjericks.github.io/py-gdalogr-cookbook/raster_layers.html#create-least-cost-path

        * costSurfaceImg - Input image to calculate cost path from
        * outputPathImg - Output image
        * startCoord - Start coordinate
        * stopCoord - End coordinate
        * gdalformat - GDAL format (default=KEA)
        * costImgBand - Band in input image to use for cost analysis (default=1)

    """
    # Read gdal raster as array.
    gdalRasterDS = gdal.Open(costSurfaceImg)
    costSurfaceArray = gdalRasterDS.GetRasterBand(costImgBand).ReadAsArray()
    
    # Creates path array
    pathArray = _createPath(CostSurfaceImg, costSurfaceArray, startCoord, stopCoord)

    # Converts path array to raster
    _array2raster(outputPathImg, CostSurfaceImg, pathArray, gdalformat)






#! /usr/bin/env python

############################################################################
# Copyright (c) 2009 Dr. Peter Bunting, Aberystwyth University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  A series of classes to generate image plots from the ptxt text 
#           file format exported for plotting from the rsgis library. 
#
# Author: Pete Bunting
# Email: pete.bunting@aber.ac.uk
# Date: 11/08/2009
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
#############################################################################

import os.path
import sys, re
import numpy as np
from pylab import *
try:
    import numpy as np
except ImportError:
    print('''Numpy is required for this script, not found.
Check: http://numpy.scipy.org/''')
    exit()
try:
    import matplotlib.pyplot as plt
except ImportError:
    print('''MatPlotLib is required for this script, not found.
Check: http://matplotlib.sourceforge.net/''')
    exit()
try:
    from enthought.mayavi import mlab
    importMlab = True
except ImportError:
    importMlab = False

def checkforHash(line):
    foundHash = False
    for i in range(len(line)):
        if line[i] == '#':
            foundHash = True
    return foundHash

def stringTokenizer(line, delimiter):
    tokens = list()
    token = str()
    for i in range(len(line)):
        if line[i] == delimiter:
            tokens.append(token)
            token = str()
        else:
            token = token + line[i]
    tokens.append(token)
    return tokens

class PlotFreq2D (object):
    
    def plot(self, data, outputFile, outFormat):
        fig = plt.figure(dpi=80)
        plt.hist(data,bins=30)
        plt.savefig(outputFile, format=outFormat)
    
    def readPTXTFile(self, inputFile, data):
        dataFile = open(inputFile, 'r')
        for eachLine in dataFile:
            comment = checkforHash(eachLine)
            if comment == False:
                data.append(float(eachLine.strip())) 
        dataFile.close()
    
    def help(self):
        plotObj = RSGISLibPlot();
        plotObj.help()
        print('Additional Parameters for Freq2D.')
        print('\t<FORMAT> - output file format of the plot (PNG | PDF)')
    
    def createPlot(self):
        print("PlotFreq2D - Basic Implementation..")
        numargs = len(sys.argv)
        if numargs == 4:
            inputFile = sys.argv[1].strip()
            outputFile = sys.argv[2].strip()
            format = sys.argv[3].strip()
            
            data = list()
            self.readPTXTFile(inputFile, data)
            self.plot(data, outputFile, format)
        else:
            self.help()
        
class PlotFreq3D (object):
    
    def createPlot(self):
        print("PlotFreq3D - NOT IMPLEMENTED YET!!")

class PlotLines2D (object):
    
    def createPlot(self):
        print("PlotLines2D - NOT IMPLEMENTED YET!!")

class PlotLines3D (object):
    
    def createPlot(self):
        print("PlotLines3D - NOT IMPLEMENTED YET!!")

class PlotDensity (object):
    
    def createPlot(self):
        print("PlotDensity - NOT IMPLEMENTED YET!!")

class Plot2DScatter (object):
    
    def plot(self, dataX, dataY, outputFile, outFormat):
        fig = plt.figure(dpi=80)
        plt.scatter(dataX, dataY, marker='o')
        plt.savefig(outputFile, format=outFormat)
    
    def readPTXTFile(self, inputFile, dataX, dataY):
        dataFile = open(inputFile, 'r')
        for eachLine in dataFile:
            comment = checkforHash(eachLine)
            if comment == False:
                tokens = stringTokenizer(eachLine, ',')
                if len(tokens) == 2:
                    dataX.append(float(tokens[0].strip()))
                    dataY.append(float(tokens[1].strip()))
                else:
                    raise Exception('Not enough tokens in a line')
        dataFile.close()
    
    def help(self):
        plotObj = RSGISLibPlot();
        plotObj.help()
        print('Additional Parameters for Freq2D.')
        print('\t<FORMAT> - output file format of the plot (PNG | PDF)')
    
    def createPlot(self):
        print("Plot2DScatter - Basic Implementation..")
        numargs = len(sys.argv)
        if numargs == 4:
            inputFile = sys.argv[1].strip()
            outputFile = sys.argv[2].strip()
            format = sys.argv[3].strip()
            
            dataX = list()
            dataY = list()
            try:
                self.readPTXTFile(inputFile, dataX, dataY)
                self.plot(dataX, dataY, outputFile, format)
            except Exception as e:
                print("Error: ", e)
        else:
            self.help()
        
class PlotClass2DScatter (object):
    def plot(self, dataX, dataY, classNames, classNamesLowLim, classNamesUpLim, outputFile, outFormat):
        fig = plt.figure(dpi=80)
                
        for i in range(len(classNames)):
            classCol = self.stdCol(i)
            className = classNames[i]
            lowLim = int(classNamesLowLim[i])
            upLim = int(classNamesUpLim[i])
            dataA = []
            dataB = []
            for j in range(upLim-lowLim):
                dataA.append(dataX[lowLim + j])
                dataB.append(dataY[lowLim + j])
            plt.scatter(dataA, dataB, marker='o',edgecolor=classCol, facecolor=classCol, label=className)
            
        plt.legend()
        plt.savefig(outputFile, format=outFormat)
    
    def readPTXTFile(self, inputFile, dataRawClass, dataRawX, dataRawY, classNames):
        dataFile = open(inputFile, 'r')
        lineNo = 0
        for eachLine in dataFile:
            if lineNo == 1: # Class names stored in second line of file
                eachLine = re.sub('#', '', eachLine)
                tokensC = stringTokenizer(eachLine, ',')
                for i in range(len(tokensC)):
                    classNames.append(str(tokensC[i].strip()))
            else:
                comment = checkforHash(eachLine)
                if comment == False:
                    tokens = stringTokenizer(eachLine, ',')
                    if len(tokens) == 3:
                        dataRawClass.append(str(tokens[0].strip()))
                        dataRawX.append(float(tokens[1].strip()))
                        dataRawY.append(float(tokens[2].strip()))
                    else:
                        raise Exception('Not enough tokens in a line')
            lineNo = lineNo + 1 
        dataFile.close()
        
    def sortData(self, dataRawClass, dataRawX, dataRawY, dataX, dataY, classNames, classNamesLowLim, classNamesUpLim):
        totalData = 0
        
        for c in range(len(classNames)):
            currentClass = classNames[c]
            numDataClass = 0
            for i in range(len(dataRawClass)):
                elementClassName = dataRawClass[i]
                if elementClassName == currentClass:
                    dataX.append(dataRawX[i])
                    dataY.append(dataRawY[i])
                    numDataClass = numDataClass + 1
                                
            classNamesLowLim.append(totalData)
            classNamesUpLim.append(totalData + numDataClass)
            
            totalData = totalData + numDataClass
    def stdCol(self, i):
        sdColScheme = list()
        sdColScheme.append("r")
        sdColScheme.append("g")
        sdColScheme.append("b")
        sdColScheme.append("c")
        sdColScheme.append("m")
        sdColScheme.append("y")
        
        if i < len(sdColScheme):
            colour = sdColScheme[i]
        else:
            colour = cm.jet(i*100)
        return colour
    
    def help(self):
        plotObj = RSGISLibPlot();
        plotObj.help()
        print('Additional Parameters for Class2DScatter.')
        print('\t<FORMAT> - output file format of the plot (PNG | PDF)')
    
    def createPlot(self):
        print("Class2DScatter ")
        numargs = len(sys.argv)
        if numargs == 4:
            inputFile = sys.argv[1].strip()
            outputFile = sys.argv[2].strip()
            format = sys.argv[3].strip()
            
            classNames = list()
            dataRawClass = list()
            dataRawX = list()
            dataRawY = list()
            
            dataX = list()
            dataY = list()
            classNamesLowLim = list()
            classNamesUpLim = list()
            
            try:
                self.readPTXTFile(inputFile, dataRawClass, dataRawX, dataRawY, classNames)
                self.sortData(dataRawClass, dataRawX, dataRawY, dataX, dataY, classNames, classNamesLowLim, classNamesUpLim)
                del dataRawClass
                del dataRawX
                del dataRawY
                self.plot(dataX, dataY, classNames, classNamesLowLim, classNamesUpLim, outputFile, format)
            except Exception as e:
                print("Error: ", e)
        else:
            self.help()

class Plot3DScatter (object):
    
    def plot(self, dataX, dataY, dataZ, outputFile, outFormat):
        dataXBaseZero = list()
        dataXMin = min(dataX);
        for item in dataX:
            dataXBaseZero.append(item - dataXMin)
        
        dataYBaseZero = list()
        dataYMin = min(dataY);
        for item in dataY:
            dataYBaseZero.append(item - dataYMin)
        
        dataXMax = max(dataXBaseZero)
        dataYMax = max(dataYBaseZero)
        ptsMaxXAxis = 0
        if dataXMax >= dataYMax:
            ptsMaxXAxis = dataXMax
        else:
            ptsMaxXAxis = dataYMax
        
        
        axisdimensionsPts = [-2,(ptsMaxXAxis+2),(min(dataZ)-2),(max(dataZ)+2)]
        
        plt.figure(dpi="200")
        
        plt.subplot(121)
        ylabel("Height")
        xlabel("X")
        plt.scatter(dataXBaseZero, dataZ, color='black', marker='o')
        axis(axisdimensionsPts)
        grid()
        
        plt.subplot(122)
        xlabel("Y")
        plt.scatter(dataYBaseZero, dataZ, color='black', marker='o')
        axis(axisdimensionsPts)
        grid()
        plt.savefig(outputFile, format=outFormat)
    
    def readPTXTFile(self, inputFile, dataX, dataY, dataZ):
        dataFile = open(inputFile, 'r')
        for eachLine in dataFile:
            comment = checkforHash(eachLine)
            if comment == False:
                tokens = stringTokenizer(eachLine, ',')
                if len(tokens) == 3:
                    dataX.append(float(tokens[0].strip()))
                    dataY.append(float(tokens[1].strip()))
                    dataZ.append(float(tokens[2].strip()))
                else:
                    raise Exception('Not enough tokens in a line')
        dataFile.close()
    
    def help(self):
        plotObj = RSGISLibPlot();
        plotObj.help()
        print('Additional Parameters for scatter3D.')
        print('\t<FORMAT> - output file format of the plot (PNG | PDF)')
    
    def createPlot(self):
        print("Plot3DScatter - Basic Implementation..")
        numargs = len(sys.argv)
        if numargs == 4:
            inputFile = sys.argv[1].strip()
            outputFile = sys.argv[2].strip()
            format = sys.argv[3].strip()
            
            dataX = list()
            dataY = list()
            dataZ = list()
            try:
                self.readPTXTFile(inputFile, dataX, dataY, dataZ)
                self.plot(dataX, dataY, dataZ, outputFile, format)
            except Exception as e:
                print("Error: ", e)
        else:
            self.help()

class Plotc2DScatter (object):

    def createPlot(self):
        print("Plotc2DScatter - NOT IMPLEMENTED YET!!")

class Plotc3DScatter (object):
    
    def createPlot(self):
        print("Plotc3DScatter - NOT IMPLEMENTED YET!!")

class PlotSurface (object):
    
    def plot(self, dataX, dataY, dataZ, outputFile):
        maxX = int(max(dataX))
        maxY = int(max(dataY))
        
        grid = list()
        for i in range(maxY):
            grid.append(list())
            for j in range(maxX):
                grid[i].append(0)
        #x = int(0)
        #y = int(0)
        for i in range(len(dataX)):
            x = dataX[i]-1
            y = dataY[i]-1
            grid[y][x]=dataZ[i]
        
        #print grid
        
        fig = mlab.figure()
        mlab.surf(grid, figure=fig)
        mlab.savefig(outputFile, figure=fig)
        mlab.show()
    
    def readPTXTFile(self, inputFile, dataX, dataY, dataZ):
        dataFile = open(inputFile, 'r')
        for eachLine in dataFile:
            comment = checkforHash(eachLine)
            if comment == False:
                tokens = stringTokenizer(eachLine, ',')
                if len(tokens) == 3:
                    dataX.append(int(tokens[0].strip()))
                    dataY.append(int(tokens[1].strip()))
                    dataZ.append(float(tokens[2].strip()))
                else:
                    raise Exception('Not enough tokens in a line')
        dataFile.close()
    
    def help(self):
        plotObj = RSGISLibPlot();
        plotObj.help()
    
    def createPlot(self):
        print("PlotSurface - Basic Implementation..")
        if importMlab == False:
            print('mayavi, from Enthough Python distribution is required for surface plots, not found')
            exit()
        numargs = len(sys.argv)
        if numargs == 4: 
            inputFile = sys.argv[1].strip()
            outputFile = sys.argv[2].strip()
            
            dataX = list()
            dataY = list()
            dataZ = list()
            try:
                print("Reading Data ")
                self.readPTXTFile(inputFile, dataX, dataY, dataZ)   
                self.plot(dataX, dataY, dataZ, outputFile)
            except Exception as e:
                print("Error: ", e)
        else:
            self.help()

class PlotcSurface (object):
    
    def createPlot(self):
        print("PlotcSurface - NOT IMPLEMENTED YET!!")


class RSGISLibPlot (object):

    def runPlot(self, inputFile):
        inputPTXTFile = open(inputFile, 'r')
        plottype = str()
        for line in inputPTXTFile:
            plottype = line.strip()
            break
        inputPTXTFile.close()
        
        if plottype == "#freq2D":
            #print "Plot freq2D"
            plotObj = PlotFreq2D()
            plotObj.createPlot()
        elif plottype == "#freq3D":
            print("Plot freq3D")
            plotObj = PlotFreq3D()
            plotObj.createPlot()
        elif plottype == "#lines2D":
            print("Plot lines2D")
            plotObj = PlotLines2D()
            plotObj.createPlot()
        elif plottype == "#lines3D":
            print("Plot lines3D")
            plotObj = PlotLines3D()
            plotObj.createPlot()
        elif plottype == "#Density":
            print("Plot Density")
            plotObj = PlotDensity()
            plotObj.createPlot()
        elif plottype == "#2DScatter":
            print("Plot 2DScatter")
            plotObj = Plot2DScatter()
            plotObj.createPlot()
        elif plottype == "#ClassScatter2D":
            print("Plot ClassScatter2D")
            plotObj = PlotClass2DScatter()
            plotObj.createPlot()
        elif plottype == "#3DScatter":
            print("Plot 3DScatter")
            plotObj = Plot3DScatter()
            plotObj.createPlot()
        elif plottype == "#c2DScatter":
            print("Plot c2DScatter")
            plotObj = Plotc2DScatter()
            plotObj.createPlot()
        elif plottype == "#c3DScatter":
            print("Plot c3DScatter")
            plotObj = Plotc3DScatter()
            plotObj.createPlot()
        elif plottype == "#Surface":
            print("Plot Surface")
            plotObj = PlotSurface()
            plotObj.createPlot()
        elif plottype == "#cSurface":
            print("Plot cSurface")
            plotObj = PlotcSurface()
            plotObj.createPlot()
        else:
            print("ERROR: Did not recognise the plot type \'" + plottype + "\'.")
            
    
    def run(self):
        numArgs = len(sys.argv)
        if numArgs > 2:
            inputFile = sys.argv[1].strip()
            
            if os.path.exists(inputFile):
                self.runPlot(inputFile)
            else:
                print('Input File does not exist')
        else:
            self.help()
    
    def help(self):
        print('rsgislibplot.py script generates plots from ptxt files generated from the ')
        print(' software library')
        print('Usage: python rsgislibplot.py <INPUT> <OUTPUT> [PLOT SPECIFIC]')
        print('\t<INPUT> - input ptxt file')
        print('\t<OUTPUT> - output plot')
        print('\nThis script was distributed with RSGISLib 3.6.14')
        print('Copyright (C) 2014 Peter Bunting and Daniel Clewley')
        print('For support please email rsgislib-support@googlegroups.com')

if __name__ == '__main__':
    obj = RSGISLibPlot()
    obj.run()



"""
The tools.filemanipulation module contains functions for manipulating and moving files around.
"""

# Import modules
import rsgislib

import glob
import os.path
import os
import shutil

def sortImgsUTM2DIRs(inputImgsDIR, fileSearchStr, outBaseDIR):
    """
A function which will sort a series of input image files which
a projected using the UTM system into individual directories per
UTM zone. Please note that the input files are moved on your system!!

Where:

* inputImgsDIR - directory where the input files are to be found.
* fileSearchStr - the wildcard search string to find files within the input directory (e.g., \*.kea).
* outBaseDIR - the output directory where the UTM folders will be created and the files copied.

"""
    rsgisUtils = rsgislib.RSGISPyUtils()
    inFiles = glob.glob(os.path.join(inputImgsDIR, fileSearchStr))
    for imgFile in inFiles:
        utmZone = rsgisUtils.getUTMZone(imgFile)
        if utmZone is not None:
            outDIR = os.path.join(outBaseDIR, 'utm'+utmZone)
            if not os.path.exists(outDIR):
                os.makedirs(outDIR)
            imgFileList = rsgisUtils.getImageFiles(imgFile)
            for tmpFile in imgFileList:
                print('Moving: ' + tmpFile)
                outFile = os.path.join(outDIR, os.path.basename(tmpFile))
                shutil.move(tmpFile, outFile)








#!/usr/bin/env python


############################################################################
#  solarangles.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose: Provide a set of utilities which provide further functionality
#          specifically for caclulating solar angles.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 23/11/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

from osgeo import ogr
from osgeo import osr
import numpy

havePysolar = True
try:
    import Pysolar
except ImportError:
    havePysolar = False

haveRIOS = True
try:
    from rios import applier
    from rios import cuiprogress
except ImportError:
    haveRIOS = False

import datetime


def getSolarIrrConventionSolarAzimuthFromUSGS(solarAz):
    """
IN: USGS Convertion::

              N (0)
              |
     W (-90)-----E (90) 
              |
       (-180) S (180)

OUT: Solar Irradiance Convertion::

              N (0)
              |
     W (270)-----E (90) 
              |
              S (180)

"""
    solarAzOut = solarAz
    if solarAz < 0:
        solarAzOut = 360 + solarAz
    return solarAzOut
    
    
def getSolarIrrConventionSolarAzimuthFromTrad(solarAz):
    """
IN: Traditional Convertion::

       (-180) N (180)
              |
     W (-90)-----E (90) 
              |
              S (0)

OUT: Solar Irradiance Convertion::

              N (0)
              |
     W (270)-----E (90) 
              |
              S (180)    

"""
    solarAzOut = 0.0
    if solarAz > 0:
        solarAzOut = 180 - solarAz
    elif solarAz < 0:
        solarAzOut = 180 + ((-1) * solarAz)
    return solarAzOut
    

def calcSolarAzimuthZenith(inputImg, inImgDateTime, outputImg, gdalformat):
    """
Function which calculate a solar azimuth (band 1) and zenith (band 2) image.

"""
    if not havePysolar:
        raise Exception("The PySolar module required for this function could not be imported.")

    if not haveRIOS:
        raise Exception("The RIOS module required for this function could not be imported.")

    infiles = applier.FilenameAssociations()
    infiles.image1 = inputImg
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = outputImg
    otherargs = applier.OtherInputs()
    otherargs.dateTime = inImgDateTime
    aControls = applier.ApplierControls()
    aControls.progress = cuiprogress.CUIProgressBar()
    aControls.drivername = gdalformat


    wgs84latlonProj = osr.SpatialReference()
    wgs84latlonProj.ImportFromEPSG(4326)
    otherargs.wgs84latlonProj = wgs84latlonProj
    
    def _calcSolarAzimuthZenith(info, inputs, outputs, otherargs):
        """
        Internal functions used within calcSolarAzimuthZenith() - don't call independently.
        
        """
        xBlock, yBlock = info.getBlockCoordArrays()
    
        inProj = osr.SpatialReference()
        inProj.ImportFromWkt(info.getProjection())
    
        transform = osr.CoordinateTransformation(inProj, otherargs.wgs84latlonProj)
        xBlockF = xBlock.flatten()
        yBlockF = yBlock.flatten()
        
        pts = numpy.zeros((xBlockF.shape[0],2), dtype=numpy.float)
        pts[...,0] = xBlockF
        pts[...,1] = yBlockF
        outPts = numpy.array(transform.TransformPoints(pts))
    
        outAz = numpy.zeros_like(xBlockF, dtype=numpy.float)
        outZen = numpy.zeros_like(xBlockF, dtype=numpy.float)
    
        for i in range(outPts.shape[0]):
            outAz[i] = 90-Pysolar.solar.GetAltitude(outPts[i,1], outPts[i,0], otherargs.dateTime)
            outZen[i] = ((-1)*Pysolar.solar.GetAzimuth(outPts[i,1], outPts[i,0], otherargs.dateTime))-180
    
        outAz = numpy.reshape(outAz, xBlock.shape)
        outZen = numpy.reshape(outZen, xBlock.shape)
    
        #print("Block End")
        outputs.outimage = numpy.stack((outAz,outZen))

    # Apply the multiply function.
    applier.apply(_calcSolarAzimuthZenith, infiles, outfiles, otherargs, controls=aControls)




#! /usr/bin/env python

############################################################################
#  classratutils.py
#
#  Copyright 2015 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 28/11/2015
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function
haveGDALPy = True
try:
    import osgeo.gdal as gdal
except ImportError as gdalErr:
    haveGDALPy = False
    
haveNumpy = True
try:
    import numpy
except ImportError as numErr:
    haveNumpy = False

haveRIOSRat = True
try:
    from rios import rat
    from rios import ratapplier
except ImportError as riosRatErr:
    haveRIOSRat = False

haveSKLearnRF = True
try:
    from sklearn.ensemble import RandomForestClassifier
except ImportError as sklearnRFErr:
    haveSKLearnRF = False
    raise Exception("The scikit-learn random forests tools are required for this module could not be imported\n\t {}".format(sklearnRFErr))
    
haveSKLearnGS = True
try:
    from sklearn.model_selection import GridSearchCV
except ImportError as sklearnGSErr:
    haveSKLearnRF = False
    raise Exception("The scikit-learn grid search tools are required for this module could not be imported\n\t" + sklearnGSErr)

haveSKLearnPreProcess = True
try:
    from sklearn.preprocessing import MaxAbsScaler
except ImportError as sklearnPreProcessErr:
    haveSKLearnPreProcess = False
    raise Exception("The scikit-learn pre-processing modules not available.\n\t" + sklearnPreProcessErr)

haveSKLearnKM = True
try:
    from sklearn.cluster import MiniBatchKMeans
except ImportError as sklearnMBKMErr:
    haveSKLearnKM = False
    raise Exception("The scikit-learn Mini Batch KMeans tools are required for this module could not be imported\n\t" + sklearnMBKMErr)
    

def classifyWithinRAT(clumpsImg, classesIntCol, classesNameCol, variables, classifier=RandomForestClassifier(n_estimators=100, max_features=3, oob_score=True, n_jobs=-1), outColInt="OutClass", outColStr="OutClassName", roiCol=None, roiVal=1, classColours=None, preProcessor=None, justFit=False):
    """
A function which will perform a classification within the RAT using a classifier from scikit-learn

* clumpsImg is the clumps image on which the classification is to be performed
* classesIntCol is the column with the training data as int values
* classesNameCol is the column with the training data as string class names
* variables is an array of column names which are to be used for the classification
* classifier is an instance of a scikit-learn classifier (e.g., RandomForests which is Default)
* outColInt is the output column name for the int class representation (Default: 'OutClass')
* outColStr is the output column name for the class names column (Default: 'OutClassName')
* roiCol is a column name for a column which specifies the region to be classified. If None ignored (Default: None)
* roiVal is a int value used within the roiCol to select a region to be classified (Default: 1)
* classColours is a python dict using the class name as the key along with arrays of length 3 specifying the RGB colours for the class.
* preProcessor is a scikit-learn processors such as sklearn.preprocessing.MaxAbsScaler() which can rescale the input variables independently as read in (Define: None; i.e., not in use).
* justFit is a boolean specifying that the classifier should just be fitted to the data and not applied (Default: False; i.e., apply classification)


Example::

    from sklearn.ensemble import ExtraTreesClassifier
    from rsgislib.classification import classratutils
    
    classifier = ExtraTreesClassifier(n_estimators=100, max_features=3, n_jobs=-1, verbose=0)
    
    classColours = dict()
    classColours['Forest'] = [0,138,0]
    classColours['NonForest'] = [200,200,200]
    
    variables = ['GreenAvg', 'RedAvg', 'NIR1Avg', 'NIR2Avg', 'NDVI']
    classifyWithinRAT(clumpsImg, classesIntCol, classesNameCol, variables, classifier=classifier, classColours=classColours)
    
    from sklearn.preprocessing import MaxAbsScaler
    
    # With pre-processor
    classifyWithinRAT(clumpsImg, classesIntCol, classesNameCol, variables, classifier=classifier, classColours=classColours, preProcessor=MaxAbsScaler())

"""
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scikit-learn RF is available
    if not haveSKLearnRF:
        raise Exception("The scikit-learn random forests tools are required for this function could not be imported\n\t" + sklearnRFErr)
    # Check scikit-learn pre-processing is available
    if not haveSKLearnPreProcess:
        raise Exception("The scikit-learn pre-processing tools are required for this function could not be imported\n\t" + sklearnPreProcessErr)
        
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    numpyVars = []
    for var in variables:
        print("Reading " + var)
        tmpArr = rat.readColumn(ratDataset, var)
        if not preProcessor is None:
            tmpArr = tmpArr.reshape(-1, 1)
            tmpArr = preProcessor.fit_transform(tmpArr)
            tmpArr = tmpArr.reshape(-1)
        numpyVars.append(tmpArr)
    
    # Read in training classes
    classesInt = rat.readColumn(ratDataset, classesIntCol)
    classesStr = rat.readColumn(ratDataset, classesNameCol)
        
    roi = None
    if not ((roiCol == None) or (roiCol == "")):
        roi = rat.readColumn(ratDataset, roiCol)
    
    # Set up output array
    outLabels = numpy.zeros_like(classesInt, dtype=numpy.int16)
    outClassNames = numpy.empty_like(classesInt, dtype=numpy.dtype('a255'))
    ID = numpy.arange(outLabels.shape[0])
    
    xData = numpy.array(numpyVars)
    xData = xData.transpose()
    xData = numpy.where(numpy.isfinite(xData), xData,0)

    print("Input data size: {} x {}".format(xData.shape[0], xData.shape[1]))
    
    trainingData = xData[numpy.isfinite(xData).all(axis=1)]
    classesInt = classesInt[numpy.isfinite(xData).all(axis=1)]
    classesStr = classesStr[numpy.isfinite(xData).all(axis=1)]
    ID = ID[numpy.isfinite(xData).all(axis=1)]
    
    trainingData = trainingData[classesInt > 0]
    classesStr = classesStr[classesInt > 0]
    classesInt = classesInt[classesInt > 0]
        
    print("Training data size: {} x {}".format(trainingData.shape[0], trainingData.shape[1]))
        
    print('Training Classifier')
    classifier.fit(trainingData, classesInt)
    
    print('Calc Classifier Accuracy')
    accVal = classifier.score(trainingData, classesInt)
    print('Classifier Score = {}'.format(round(accVal*100, 2)))
    
    if not justFit:
        if not roi is None:
            xData = xData[roi == roiVal]
            ID = ID[roi == roiVal]
            print("ROI Subsetted data size: {} x {}".format(xData.shape[0], xData.shape[1]))
        
        predClass = classifier.predict(xData)
        
        outLabels[ID] = predClass
        
        print("Writing Columns")
        rat.writeColumn(ratDataset, outColInt, outLabels)
        
        print("Create and Write Output Class Names")
        classNames = numpy.unique(classesStr)
        classes = numpy.zeros_like(classNames, dtype=numpy.int16)
        
      
        i = 0
        classNameIDs = dict()
        for className in classNames:
            classNameStr = str(className.decode())
            if not classNameStr is '':
                classes[i] = classesInt[classesStr == className][0]
                classNameIDs[classNameStr] = classes[i]
                print("Class \'" + classNameStr + "\' has numerical " + str(classes[i]))  
                i = i + 1
        
        outClassNames[...] = ''
        for className in classNameIDs:
            classID = classNameIDs[className]
            outClassNames[outLabels==classID] = className
        
        rat.writeColumn(ratDataset, outColStr, outClassNames)
        
        if not classColours is None:
            print("Set Colours")
            red = rat.readColumn(ratDataset, "Red")
            green = rat.readColumn(ratDataset, "Green")
            blue = rat.readColumn(ratDataset, "Blue")
            
            # Set Background to black
            red[...] = 0
            green[...] = 0
            blue[...] = 0
            
            # Set colours
            for className in classNameIDs:
                print("Colouring class " + className)
                classID = classNameIDs[className]
                colours = classColours[className]
                
                red   = numpy.where(outLabels == classID, colours[0], red)
                green = numpy.where(outLabels == classID, colours[1], green)
                blue  = numpy.where(outLabels == classID, colours[2], blue)
        
            rat.writeColumn(ratDataset, "Red", red)
            rat.writeColumn(ratDataset, "Green", green)
            rat.writeColumn(ratDataset, "Blue", blue)
    
    ratDataset = None



def _extractTrainDataFromRAT(info, inputs, outputs, otherargs):
    """
    This function is used internally within classifyWithinRATTiled using the RIOS ratapplier function
    """
    progress = round((info.startrow / info.rowCount)*100)
    print("{} %".format(progress), end="\r")
    
    numpyVars = []
    for var in otherargs.vars:
        varVals = getattr(inputs.inrat, var)
        numpyVars.append(varVals)
    
    xData = numpy.array(numpyVars)
    xData = xData.transpose()
    
    classIntVals = getattr(inputs.inrat, otherargs.classIntCol)
    trainingData = xData[classIntVals > 0]
    
    otherargs.trainData[otherargs.trainDataOff:otherargs.trainDataOff+trainingData.shape[0]] = trainingData
    otherargs.trainDataOff += trainingData.shape[0]

def _applyClassifier(info, inputs, outputs, otherargs):
    """
    This function is used internally within classifyWithinRATTiled using the RIOS ratapplier function
    """
    progress = round((info.startrow / info.rowCount)*100)
    print("{} %".format(progress), end="\r")    
    numpyVars = []
    for var in otherargs.vars:
        varVals = getattr(inputs.inrat, var)
        numpyVars.append(varVals)
    
    xData = numpy.array(numpyVars)
    xData = xData.transpose()
    
    ID = numpy.arange(xData.shape[0])
    outClassIntVals = numpy.zeros(xData.shape[0], dtype=numpy.int16)
    outClassNamesVals = numpy.empty(xData.shape[0], dtype=numpy.dtype('a255'))
    outClassNamesVals[...] = ''
    
    ID = ID[numpy.isfinite(xData).all(axis=1)]
    vData = xData[numpy.isfinite(xData).all(axis=1)]
        
    if not otherargs.roiCol is None:
        roi = getattr(inputs.inrat, otherargs.roiCol)
        roi = roi[numpy.isfinite(xData).all(axis=1)]
        vData = vData[roi == otherargs.roiVal]
        ID = ID[roi == otherargs.roiVal]
        #print("ROI Subsetted data size: {} x {}".format(vData.shape[0], vData.shape[1]))
        
    predClass = otherargs.classifier.predict(vData)
    
    outClassIntVals[ID] = predClass
    setattr(outputs.outrat, otherargs.outColInt, outClassIntVals)
    
    for className in otherargs.classNameIDs:
        classID = otherargs.classNameIDs[className]
        outClassNamesVals[outClassIntVals==classID] = className
    setattr(outputs.outrat, otherargs.outColStr, outClassNamesVals)
    
    if not otherargs.classColours is None:
        red = getattr(inputs.inrat, "Red")
        green = getattr(inputs.inrat, "Green")
        blue = getattr(inputs.inrat, "Blue")
        
        # Set Background to black
        red[...] = 0
        green[...] = 0
        blue[...] = 0
        
        # Set colours
        for className in otherargs.classNameIDs:
            #print("Colouring class " + className)
            classID = otherargs.classNameIDs[className]
            colours = otherargs.classColours[className]
            
            red   = numpy.where(outClassIntVals == classID, colours[0], red)
            green = numpy.where(outClassIntVals == classID, colours[1], green)
            blue  = numpy.where(outClassIntVals == classID, colours[2], blue)
    
        setattr(outputs.outrat, "Red", red)
        setattr(outputs.outrat, "Green", green)
        setattr(outputs.outrat, "Blue", blue)
    
    

def classifyWithinRATTiled(clumpsImg, classesIntCol, classesNameCol, variables, classifier=RandomForestClassifier(n_estimators=100, max_features=3, oob_score=True, n_jobs=-1), outColInt="OutClass", outColStr="OutClassName", roiCol=None, roiVal=1, classColours=None, scaleVarsRange=False, justFit=False):
    """
A function which will perform a classification within the RAT using a classifier from scikit-learn using the rios ratapplier interface allowing very large RATs to be processed. 

* clumpsImg is the clumps image on which the classification is to be performed
* classesIntCol is the column with the training data as int values
* classesNameCol is the column with the training data as string class names
* variables is an array of column names which are to be used for the classification
* classifier is an instance of a scikit-learn classifier (e.g., RandomForests which is Default)
* outColInt is the output column name for the int class representation (Default: 'OutClass')
* outColStr is the output column name for the class names column (Default: 'OutClassName')
* roiCol is a column name for a column which specifies the region to be classified. If None ignored (Default: None)
* roiVal is a int value used within the roiCol to select a region to be classified (Default: 1)
* classColours is a python dict using the class name as the key along with arrays of length 3 specifying the RGB colours for the class.
* scaleVarsRange will rescale each variable independently to a range of 0-1 (default: False).
* justFit is a boolean specifying that the classifier should just be fitted to the data and not applied (Default: False; i.e., apply classification)


Example::

    from sklearn.ensemble import ExtraTreesClassifier
    from rsgislib.classification import classratutils
    
    classifier = ExtraTreesClassifier(n_estimators=100, max_features=3, n_jobs=-1, verbose=0)
    
    classColours = dict()
    classColours['Forest'] = [0,138,0]
    classColours['NonForest'] = [200,200,200]
    
    variables = ['GreenAvg', 'RedAvg', 'NIR1Avg', 'NIR2Avg', 'NDVI']
    classifyWithinRATTiled(clumpsImg, classesIntCol, classesNameCol, variables, classifier=classifier, classColours=classColours)
        
    # With using range scaling.
    classifyWithinRATTiled(clumpsImg, classesIntCol, classesNameCol, variables, classifier=classifier, classColours=classColours, scaleVarsRange=True)

"""
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scikit-learn RF is available
    if not haveSKLearnRF:
        raise Exception("The scikit-learn random forests tools are required for this function could not be imported\n\t" + sklearnRFErr)
    # Check scikit-learn pre-processing is available
    if not haveSKLearnPreProcess:
        raise Exception("The scikit-learn pre-processing tools are required for this function could not be imported\n\t" + sklearnPreProcessErr)
        
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    
    # Read in training classes
    classesInt = rat.readColumn(ratDataset, classesIntCol)
    classesStr = rat.readColumn(ratDataset, classesNameCol)
    ratDataset = None
    
    validClassStr = classesStr[classesInt > 0]
    validClassInt = classesInt[classesInt > 0]
    
    #print(validClassInt.shape)
    classNames = numpy.unique(validClassStr)
    classes = numpy.zeros_like(classNames, dtype=numpy.int16)
    
    i = 0
    classNameIDs = dict()
    for className in classNames:
        classNameStr = str(className.decode())
        if not classNameStr is '':
            #print(validClassInt[validClassStr == className])
            classes[i] = validClassInt[validClassStr == className][0]
            classNameIDs[classNameStr] = classes[i]
            #print("Class \'" + classNameStr + "\' has numerical " + str(classes[i]))  
            i = i + 1    
    
    trainLen = validClassInt.shape[0]
    numVars = len(variables)
    
    #print("Create numpy {} x {} array for training".format(trainLen, numVars))
    
    trainData = numpy.zeros((trainLen, numVars), numpy.float64)    
    
    in_rats = ratapplier.RatAssociations()
    out_rats = ratapplier.RatAssociations()
    in_rats.inrat = ratapplier.RatHandle(clumpsImg)

    otherargs = ratapplier.OtherArguments()
    otherargs.vars = variables
    otherargs.classIntCol = classesIntCol
    otherargs.trainData = trainData
    otherargs.trainDataOff = 0
    
    print("Extract Training Data")
    ratapplier.apply(_extractTrainDataFromRAT, in_rats, out_rats, otherargs=otherargs, controls=None)
    print("100%")
    
    
    validClassInt = validClassInt[numpy.isfinite(trainData).all(axis=1)]
    validClassStr = validClassStr[numpy.isfinite(trainData).all(axis=1)]
    trainData = trainData[numpy.isfinite(trainData).all(axis=1)]
    
    print("Training data size: {} x {}".format(trainData.shape[0], trainData.shape[1]))
    
    print('Training Classifier')
    classifier.fit(trainData, validClassInt)
    print("Completed")
    
    print('Calc Classifier Accuracy')
    accVal = classifier.score(trainData, validClassInt)
    print('Classifier Score = {}'.format(round(accVal*100, 2)))
    
    if not justFit:
        print("Apply Classifier")
        in_rats = ratapplier.RatAssociations()
        out_rats = ratapplier.RatAssociations()
        in_rats.inrat = ratapplier.RatHandle(clumpsImg)
        out_rats.outrat = ratapplier.RatHandle(clumpsImg)
        
        otherargs = ratapplier.OtherArguments()
        otherargs.vars = variables
        otherargs.classifier = classifier
        otherargs.outColInt = outColInt
        otherargs.outColStr = outColStr
        otherargs.roiCol = roiCol
        otherargs.roiVal = roiVal
        otherargs.classColours = classColours
        otherargs.classNameIDs = classNameIDs
        
        
        ratapplier.apply(_applyClassifier, in_rats, out_rats, otherargs=otherargs, controls=None)
        print("100%")





def clusterWithinRAT(clumpsImg, variables, clusterer=MiniBatchKMeans(n_clusters=8, init='k-means++', max_iter=100, batch_size=100), outColInt="OutCluster", roiCol=None, roiVal=1, clrClusters=True, clrSeed=10, addConnectivity=False, preProcessor=None):
    """
A function which will perform a clustering within the RAT using a clustering algorithm from scikit-learn

* clumpsImg is the clumps image on which the classification is to be performed.
* variables is an array of column names which are to be used for the clustering.
* clusterer is an instance of a scikit-learn clusterer (e.g., MiniBatchKMeans which is Default; Note with 8 clusters).
* outColInt is the output column name identifying the clusters (Default: 'OutCluster').
* roiCol is a column name for a column which specifies the region to be clustered. If None ignored (Default: None).
* roiVal is a int value used within the roiCol to select a region to be clustered (Default: 1).
* clrClusters is a boolean specifying whether the colour table should be updated to correspond to the clusters (Default: True).
* clrSeed is an integer seeding the random generator used to generate the colours (Default=10; if None provided system time used).
* addConnectivity is a boolean which adds a kneighbors_graph to the clusterer (just an option for the AgglomerativeClustering algorithm)
* preProcessor is a scikit-learn processors such as sklearn.preprocessing.MaxAbsScaler() which can rescale the input variables independently as read in (Define: None; i.e., not in use).


Example::

    from rsgislib.classification import classratutils
    from sklearn.cluster import DBSCAN
    
    sklearnClusterer = DBSCAN(eps=1, min_samples=50)
    classratutils.clusterWithinRAT('MangroveClumps.kea', ['MinX', 'MinY'], clusterer=sklearnClusterer, outColInt="OutCluster", roiCol=None, roiVal=1, clrClusters=True, clrSeed=10, addConnectivity=False)
    
    # With pre-processor
    from sklearn.preprocessing import MaxAbsScaler
    classratutils.clusterWithinRAT('MangroveClumps.kea', ['MinX', 'MinY'], clusterer=sklearnClusterer, outColInt="OutCluster", roiCol=None, roiVal=1, clrClusters=True, clrSeed=10, addConnectivity=False, preProcessor=MaxAbsScaler())

"""
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scikit-learn RF is available
    if not haveSKLearnKM:
        raise Exception("The scikit-learn Mini Batch KMeans tools are required for this function could not be imported\n\t" + sklearnMBKMErr)
    # Check scikit-learn pre-processing is available
    if not haveSKLearnPreProcess:
        raise Exception("The scikit-learn pre-processing tools are required for this function could not be imported\n\t" + sklearnPreProcessErr)
        
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    Histogram = rat.readColumn(ratDataset, 'Histogram')
    numpyVars = []
    for var in variables:
        print("Reading " + var)
        tmpArr = rat.readColumn(ratDataset, var)
        if not preProcessor is None:
            tmpArr = tmpArr.reshape(-1, 1)
            tmpArr = preProcessor.fit_transform(tmpArr)
            tmpArr = tmpArr.reshape(-1)
        numpyVars.append(tmpArr)
        
    roi = None
    if not ((roiCol == None) or (roiCol == "")):
        roi = rat.readColumn(ratDataset, roiCol)
    
    # Set up output array
    outLabels = numpy.zeros_like(Histogram, dtype=numpy.int16)
    ID = numpy.arange(outLabels.shape[0])
    
    xData = numpy.array(numpyVars)
    xData = xData.transpose()
    ID = ID[numpy.isfinite(xData).all(axis=1)]
    if not roi is None:
        roi = roi[numpy.isfinite(xData).all(axis=1)]
    xData = xData[numpy.isfinite(xData).all(axis=1)]
    
    if not roi is None:
        xData = xData[roi == roiVal]
        ID = ID[roi == roiVal]    
    
    print("Input Data Size: {} x {}".format(xData.shape[0], xData.shape[1]))
    
    if addConnectivity:
        from sklearn.neighbors import kneighbors_graph
        inConnectivity = kneighbors_graph(xData, n_neighbors=10, include_self=False)
        clusterer.set_params(**{'connectivity': inConnectivity})
        
    print('Fit Clusterer')
    outClust = clusterer.fit_predict(xData)    
    
    minClusterID = numpy.min(outClust)
    if minClusterID <= 0:
        minOff = 1 - minClusterID
        outClust = outClust + minOff
    
    outLabels[ID] = outClust
    
    print("Writing Columns")
    rat.writeColumn(ratDataset, outColInt, outLabels)
    
    print("Create and Write Output Class Names")
    clustersIDs = numpy.unique(outClust)
      
    if clrClusters:
        import random
        random.seed(clrSeed)
    
        print("Set Colours")
        red = rat.readColumn(ratDataset, "Red")
        green = rat.readColumn(ratDataset, "Green")
        blue = rat.readColumn(ratDataset, "Blue")
        
        # Set Background to black
        red[...] = 0
        green[...] = 0
        blue[...] = 0
        
        # Set colours
        for clusterID in clustersIDs:
            print("Colouring cluster: " + str(clusterID))
            
            red   = numpy.where(outLabels == clusterID, random.randint(0,255), red)
            green = numpy.where(outLabels == clusterID, random.randint(0,255), green)
            blue  = numpy.where(outLabels == clusterID, random.randint(0,255), blue)
    
        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)

    ratDataset = None


def findClassifierParameters(clumpsImg, classesIntCol, variables, preProcessor=None, gridSearch=GridSearchCV(RandomForestClassifier(), {})):
    """
Find the optimal parameters for a classifier using a grid search and return a classifier instance with those optimal parameters.

* clumpsImg is the clumps image on which the classification is to be performed
* classesIntCol is the column with the training data as int values
* variables is an array of column names which are to be used for the classification
* preProcessor is a scikit-learn processors such as sklearn.preprocessing.MaxAbsScaler() which can rescale the input variables independently as read in (Define: None; i.e., not in use).
* gridSearch is an instance of GridSearchCV parameterised with a classifier and parameters to be searched.

return:

* Instance of the classifier with optimal parameters defined.

Example::

    from rsgislib.classification import classratutils
    from sklearn.svm import SVC
    from sklearn.model_selection import GridSearchCV
    from sklearn.preprocessing import MaxAbsScaler
    
    clumpsImg = "./LS8_20150621_lat10lon652_r67p233_clumps.kea"
    classesIntCol = 'ClassInt'
    
    classParameters = {'kernel':['linear', 'rbf',  'poly', 'sigmoid'], 'C':[1, 2, 3, 4, 5, 10, 100, 400, 500, 1e3, 5e3, 1e4, 5e4, 1e5], 'gamma':[0.0001, 0.0005, 0.001, 0.005, 0.01, 0.1, 'auto'], 'degree':[2, 3, 4, 5, 6, 7, 8], 'class_weight':['', 'balanced'], 'decision_function_shape':['ovo', 'ovr', None]}
    variables = ['BlueRefl', 'GreenRefl', 'RedRefl', 'NIRRefl', 'SWIR1Refl', 'SWIR2Refl']
    
    gSearch = GridSearchCV(SVC(), classParameters)
    classifier = classratutils.findClassifierParameters(clumpsImg, classesIntCol, variables, preProcessor=MaxAbsScaler(), gridSearch=gSearch)

"""
    # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    # Check scikit-learn pre-processing is available
    if not haveSKLearnPreProcess:
        raise Exception("The scikit-learn pre-processing tools are required for this function could not be imported\n\t" + sklearnPreProcessErr)
    # Check scikit-learn Grid Search is available
    if not haveSKLearnGS:
        raise Exception("The scikit-learn grid search tools are required for this function could not be imported\n\t" + sklearnGSErr)
        
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    numpyVars = []
    for var in variables:
        print("Reading " + var)
        tmpArr = rat.readColumn(ratDataset, var)
        if not preProcessor is None:
            tmpArr = tmpArr.reshape(-1, 1)
            tmpArr = preProcessor.fit_transform(tmpArr)
            tmpArr = tmpArr.reshape(-1)
        numpyVars.append(tmpArr)
    
    # Read in training classes
    classesInt = rat.readColumn(ratDataset, classesIntCol)
        
    xData = numpy.array(numpyVars)
    xData = xData.transpose()
    xData = numpy.where(numpy.isfinite(xData), xData,0)

    print("Input data size: {} x {}".format(xData.shape[0], xData.shape[1]))
    
    trainingData = xData[numpy.isfinite(xData).all(axis=1)]
    classesInt = classesInt[numpy.isfinite(xData).all(axis=1)]
    
    trainingData = trainingData[classesInt > 0]
    classesInt = classesInt[classesInt > 0]
        
    print("Training data size: {} x {}".format(trainingData.shape[0], trainingData.shape[1]))
    print("Training data IDs size: {}".format(classesInt.shape[0]))
    
    classIDs = numpy.unique(classesInt)
    print(classIDs)
    for id in classIDs:
        print("Class {} has {} samples.".format(id, classesInt[classesInt==id].shape[0]))
    
    
    gridSearch.fit(trainingData, classesInt)
    if not gridSearch.refit:
        raise Exception("Grid Search did no find a fit therefore failed...")
    
    print("Best score was {} and has parameters {}.".format(gridSearch.best_score_, gridSearch.best_params_))
   
    return gridSearch.best_estimator_


def balanceSampleTrainingRandom(clumpsImg, trainCol, outTrainCol, minNoSamples, maxNoSamples):
    """
A function to balance the number of training samples for classification so the number is above
a minimum threshold (minNoSamples) and all equal to the class with the smallest number of samples
unless that is above a set maximum (maxNoSamples).

* clumpsImg is a string with the file path to the input image with RAT
* trainCol is a string for the name of the input column specifying the training samples (zero is no data)
* outTrainCol is a string with the name of the outputted training samples.
* minNoSamples is an int specifying the minimum number of training samples for a class (if below threshold class is removed).
* maxNoSamples is an int specifiying the maximum number of training samples per class.

"""
     # Check gdal is available
    if not haveGDALPy:
        raise Exception("The GDAL python bindings required for this function could not be imported\n\t" + gdalErr)
    # Check numpy is available
    if not haveNumpy:
        raise Exception("The numpy module is required for this function could not be imported\n\t" + numErr)
    # Check rios rat is available
    if not haveRIOSRat:
        raise Exception("The RIOS rat tools are required for this function could not be imported\n\t" + riosRatErr)
    
    ratDataset = gdal.Open(clumpsImg, gdal.GA_Update)
    trainColVals = rat.readColumn(ratDataset, trainCol)
    trainColOutVals = numpy.zeros_like(trainColVals)
    
    classIDs = numpy.unique(trainColVals)
    classIDs = classIDs[classIDs != 0]
    
    numSampPerClass = []    
    print("Number of input samples:")
    for id in classIDs:
        numVals = trainColVals[trainColVals==id].shape[0]
        print("\tClass {} has {} samples.".format(id, numVals))
        numSampPerClass.append(numVals)
    
    minNumSamples = 0
    first = True
    for i in range(len(numSampPerClass)):
        if numSampPerClass[i] < minNoSamples:
            trainColOutVals[trainColVals == classIDs[i]] = 0
        else:
            if first:
                minNumSamples = numSampPerClass[i]
                first = False
            elif numSampPerClass[i] < minNumSamples:
                minNumSamples = numSampPerClass[i]
        
    if minNumSamples > maxNoSamples:
        minNumSamples = maxNoSamples
    
    print("Number of output samples:") 
    for i in range(len(numSampPerClass)):
        if numSampPerClass[i] >= minNoSamples:
            indexes = numpy.where(trainColVals == classIDs[i])
            sampleIdx = numpy.random.choice(indexes[0], minNumSamples, replace=False)
            trainColOutVals[sampleIdx] = classIDs[i]
        print("\tClass {} has {} samples.".format(classIDs[i], trainColOutVals[trainColOutVals==classIDs[i]].shape[0]))
    
    rat.writeColumn(ratDataset, outTrainCol, trainColOutVals)
    
    ratDataset = None




"""
The classification module provides classification functionality within RSGISLib.
"""

# import the C++ extension into this level
from ._classification import *


# Import the RSGISLib Image Utils module
from rsgislib import vectorutils

def generateTransectAccuracyPts(inputImage, inputLinesShp, outputPtsShp, classImgCol, classImgVecCol, classRefVecCol, lineStep, force=False):
    """ A tool for converting a set of lines in to point transects and 
populating with the information for undertaking an accuracy assessment.

Where:

* inputImage is a string specifying the input image file with classification.
* inputLinesShp is a string specifying the input lines shapefile path. 
* outputPtsShp is a string specifying the output points shapefile path.
* classImgCol is a string speciyfing the name of the column in the image file containing the class names.
* classImgVecCol is a string specifiying the output column in the shapefile for the classified class names.
* classRefVecCol is an optional string specifiying an output column in the shapefile which can be used in the accuracy assessment for the reference data.
* lineStep is a double specifying the step along the lines between the points
* force is an optional boolean specifying whether the output shapefile should be deleted if is already exists (True and it will be deleted; Default is False)
    """
    vectorutils.createLinesOfPoints(inputLinesShp, outputPtsShp, lineStep, force)
    
    popClassInfoAccuracyPts(inputImage, outputPtsShp, classImgCol, classImgVecCol, classRefVecCol)
    
    
    


"""
The Raster GIS module contains functions for attributing and manipulating raster attribute tables.
"""

# import the C++ extension into this level
from ._rastergis import *

haveGDALPy = True
try:
    import osgeo.gdal as gdal
except ImportError as gdalErr:
    haveGDALPy = False

haveNumpy = True
try:
    import numpy
except ImportError as numErr:
    haveNumpy = False

haveHDF5 = True
try:
    import h5py
except ImportError as h5Err:
    haveHDF5 = False


class BandAttStats:
    """ This is passed to the populateRATWithStats function """
    def __init__(self, band, minField=None, maxField=None, sumField=None, stdDevField=None, meanField=None):
        self.band = band
        self.minField = minField
        self.maxField = maxField
        self.sumField = sumField
        self.meanField = meanField
        self.stdDevField = stdDevField
        
class FieldAttStats:
    """ This is passed to the calcRelDiffNeighStats function """
    def __init__(self, field, minField=None, maxField=None, sumField=None, stdDevField=None, meanField=None):
        self.field = field
        self.minField = minField
        self.maxField = maxField
        self.sumField = sumField
        self.meanField = meanField
        self.stdDevField = stdDevField

class BandAttPercentiles:
    """ This is passed to the populateRATWithPercentiles function """
    def __init__(self, percentile, fieldName):
        self.percentile = percentile
        self.fieldName = fieldName

class ShapeIndex:
    """ This is passed to the calcShapeIndices function """
    def __init__(self, colName, idx, colIdx=0):
        self.colName = colName
        self.colIdx = colIdx
        self.idx = idx

def exportCols2GDALImage(clumps, outimage, gdalformat, datatype, fields, ratband=1, tempDIR=None):
    """Exports columns of the raster attribute table as bands in a GDAL image. Utility function, exports each column individually then stacks them.

Where:

* clumps is a string containing the name of the input image file with RAT
* outimage is a string containing the name of the output gdal file
* gdalformat is a string containing the GDAL format for the output file - eg 'KEA'
* datatype is an int containing one of the values from rsgislib.TYPE_*
* field is a list of strings, providing the names of the column to be exported.
* ratband is an optional (default = 1) integer parameter specifying the image band to which the RAT is associated.

Example::

   clumps='./RATS/injune_p142_casi_sub_utm_clumps_elim_final_clumps_elim_final.kea'
   outimage='./TestOutputs/RasterGIS/injune_p142_casi_rgb_export.kea'
   gdalformat = 'KEA'
   datatype = rsgislib.TYPE_32FLOAT
   fields = ['RedAvg','GreenAvg','BlueAvg']
   rastergis.exportCols2GDALImage(clumps, outimage, gdalformat, datatype, fields)

"""
    
    import os
    import rsgislib
    from rsgislib import imageutils

    rsgisUtils = rsgislib.RSGISPyUtils()

    if tempDIR is None:
        tempDIR = os.path.split(outimage)[0]

    outExt=os.path.splitext(outimage)[-1]
    tempFileList = []

    # Export each field
    for field in fields:
        print('Exporting: ' + field)
        outTempFile = os.path.join(tempDIR, field + outExt)
        exportCol2GDALImage(clumps, outTempFile, gdalformat, datatype, field, ratband=1)
        tempFileList.append(outTempFile)

    # Stack Bands
    print('Stacking Bands')
    imageutils.stackImageBands(tempFileList, fields, outimage, None, 0, gdalformat, datatype)

    # Remove temp files
    print('Removing temp files')
    for tempFile in tempFileList:
        rsgisUtils.deleteFileWithBasename(tempFile)











"""
The tools.visualisation module contains functions for aiding visualisation of data.
"""

# Import modules
import rsgislib
import rsgislib.imageutils

import os.path
import os
import shutil
import subprocess 


def createKMZImg(inputImg, outputFile, bands, reprojLatLong=True, finiteMsk=False):
    """
    A function to convert an input image to a KML/KMZ file, where the input image
    is stretched and bands sub-selected / ordered as required for visualisation.
    
    Where:
    
    * inputImg - input image file (any format that gdal supports)
    * outputFile - output image file (extension kmz for KMZ output / kml for KML output)
    * bands - a string (comma seperated) with the bands to be selected. (e.g., '1', '1,2,3', '5,6,4')
    * reprojLatLong - specify whether the image should be explicitly reprojected to WGS84 Lat/Long before transformation to KML.
    * finiteMsk - specify whether the image data should be masked so all values are finite before stretching.
    
    """
    
    bandLst = bands.split(',')
    multiBand = False
    if len(bandLst) == 3:
        multiBand = True
    elif len(bandLst) == 1:
        multiBand = False
    else:
        print(bandLst)
        raise rsgislib.RSGISPyException('You need to either provide 1 or 3 bands.')
    rsgisUtils = rsgislib.RSGISPyUtils()
    nImgBands = rsgisUtils.getImageBandCount(inputImg)
    
    tmpDIR = os.path.join(os.path.dirname(inputImg), rsgisUtils.uidGenerator())
    os.makedirs(tmpDIR)
    baseName = os.path.splitext(os.path.basename(inputImg))[0]
    
    selImgBandsImg = ''
    if (nImgBands == 1) and (not multiBand):
        selImgBandsImg = inputImg
    elif (nImgBands == 3) and (multiBand) and (bandLst[0] == '1') and (bandLst[1] == '2') and (bandLst[2] == '3'):
        selImgBandsImg = inputImg
    else:
        sBands = []
        for strBand in bandLst:
            sBands.append(int(strBand))
        selImgBandsImg = os.path.join(tmpDIR, baseName+'_sband.kea')
        rsgislib.imageutils.selectImageBands(inputImg, selImgBandsImg, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputImg), sBands)
    
    img2Stch = selImgBandsImg
    if finiteMsk:
        finiteMskImg = os.path.join(tmpDIR, baseName+'_FiniteMsk.kea')
        rsgislib.imageutils.genFiniteMask(selImgBandsImg, finiteMskImg, 'KEA')
        img2Stch = os.path.join(tmpDIR, baseName+'_Msk2FiniteRegions.kea')
        rsgislib.imageutils.maskImage(selImgBandsImg, finiteMskImg, img2Stch, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputImg), 0, 0)
    
    stretchImg = os.path.join(tmpDIR, baseName+'_stretch.kea')
    rsgislib.imageutils.stretchImage(img2Stch, stretchImg, False, '', True, False, 'KEA', rsgislib.TYPE_8UINT, rsgislib.imageutils.STRETCH_LINEARSTDDEV, 2)
    
    gdalInFile = stretchImg
    if reprojLatLong:
        latLongImg = os.path.join(tmpDIR, baseName+'_latlong.kea')
        outWKT = os.path.join(tmpDIR, baseName+'_latlong.wkt')
        rsgisUtils.writeList2File(['GEOGCS["WGS_1984",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.2572235630016],TOWGS84[0,0,0,0,0,0,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],AUTHORITY["EPSG","4326"]]'], outWKT)
        rsgislib.imageutils.reprojectImage(stretchImg, latLongImg, outWKT, gdalformat='KEA', interp='cubic', inWKT=None, noData=0.0, outPxlRes='auto', snap2Grid=True)
        gdalInFile = latLongImg
    
    cmd = 'gdal_translate -of KMLSUPEROVERLAY ' + gdalInFile + ' ' + outputFile
    print(cmd)
    try:
        subprocess.call(cmd, shell=True)
    except OSError as e:
       raise rsgislib.RSGISPyException('Could not execute command: ' + cmd)
        
    shutil.rmtree(tmpDIR)


def createWebTilesImg(inputImg, outputDIR, bands, zoomLevels='2-10', resample='average', finiteMsk=False):
    """
    A function to convert an input image to a tile cache for web map servers, where the input image
    is stretched and bands sub-selected / ordered as required for visualisation.
    
    Where:
    
    * inputImg - input image file (any format that gdal supports)
    * outputDIR - output directory within which the cache will be created.
    * bands - a string (comma seperated) with the bands to be selected. (e.g., '1', '1,2,3', '5,6,4')
    * reprojLatLong - specify whether the image should be explicitly reprojected to WGS84 Lat/Long before transformation to KML.
    * finiteMsk - specify whether the image data should be masked so all values are finite before stretching.
    * resample - Method of resampling (average,near,bilinear,cubic,cubicspline,lanczos,antialias)
    
    """
    
    bandLst = bands.split(',')
    multiBand = False
    if len(bandLst) == 3:
        multiBand = True
    elif len(bandLst) == 1:
        multiBand = False
    else:
        print(bandLst)
        raise rsgislib.RSGISPyException('You need to either provide 1 or 3 bands.')
    rsgisUtils = rsgislib.RSGISPyUtils()
    nImgBands = rsgisUtils.getImageBandCount(inputImg)
    
    tmpDIR = os.path.join(os.path.dirname(inputImg), rsgisUtils.uidGenerator())
    os.makedirs(tmpDIR)
    baseName = os.path.splitext(os.path.basename(inputImg))[0]
    
    selImgBandsImg = ''
    if (nImgBands == 1) and (not multiBand):
        selImgBandsImg = inputImg
    elif (nImgBands == 3) and (multiBand) and (bandLst[0] == '1') and (bandLst[1] == '2') and (bandLst[2] == '3'):
        selImgBandsImg = inputImg
    else:
        sBands = []
        for strBand in bandLst:
            sBands.append(int(strBand))
        selImgBandsImg = os.path.join(tmpDIR, baseName+'_sband.kea')
        rsgislib.imageutils.selectImageBands(inputImg, selImgBandsImg, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputImg), sBands)
    
    img2Stch = selImgBandsImg
    if finiteMsk:
        finiteMskImg = os.path.join(tmpDIR, baseName+'_FiniteMsk.kea')
        rsgislib.imageutils.genFiniteMask(selImgBandsImg, finiteMskImg, 'KEA')
        img2Stch = os.path.join(tmpDIR, baseName+'_Msk2FiniteRegions.kea')
        rsgislib.imageutils.maskImage(selImgBandsImg, finiteMskImg, img2Stch, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputImg), 0, 0)
    
    stretchImg = os.path.join(tmpDIR, baseName+'_stretch.kea')
    rsgislib.imageutils.stretchImage(img2Stch, stretchImg, False, '', True, False, 'KEA', rsgislib.TYPE_8UINT, rsgislib.imageutils.STRETCH_LINEARSTDDEV, 2)
    
    cmd = 'gdal2tiles.py -r ' + resample + ' -z ' + zoomLevels + ' -a  0 ' + stretchImg + ' ' + outputDIR
    print(cmd)
    try:
        subprocess.call(cmd, shell=True)
    except OSError as e:
       raise rsgislib.RSGISPyException('Could not execute command: ' + cmd)
        
    shutil.rmtree(tmpDIR)